from .vis.mail_sections import (
    AttachmentSection,
    AuthorSection,
    CalloutSection,
    CodeSection,
    MarkdownSection,
    ParagraphSection,
    PreviewSection,
    VisSection,
)

__all__ = [
    "VisSection",
    "PreviewSection",
    "CalloutSection",
    "AuthorSection",
    "ParagraphSection",
    "AttachmentSection",
    "MarkdownSection",
    "CodeSection",
]



from inspect import Parameter, signature
from typing import Any, Dict, List, Optional

from .plot import Plot


class NovemEmailSection(object):
    _type = "base"
    _subtype = "double"

    def __init__(self) -> None:
        """Novem e-mail sections add additional features to your e-mail."""

        self._params: List[str] = []
        self._kwparams: List[str] = []
        self._cparams: List[str] = []

        self._body: str = ""

    def get_markdown(self) -> str:
        """
        Returns novem e-mail markdown for this section
        """

        return ""


class NovemEmailSectionApi(NovemEmailSection):
    def __init__(
        self,
        locs: Dict[str, Any],
        params: Any,
        /,
        **kwargs: str,
    ) -> None:
        """Novem e-mail sections add additional features to your e-mail."""

        super().__init__()

        # get our arguments
        # args = signature(VisSection.__init__).parameters

        # construct all our keyword arguments
        for k, p in params.items():
            if p.kind != Parameter.POSITIONAL_OR_KEYWORD:
                continue

            v = locs[k]

            # #find out if k is string, bool or list
            if p.annotation is str or p.annotation is Optional[str]:
                # This is a string argument, if value is not None add it
                if v:
                    ts = f"{k.replace('_',' ')}: {v}"
                    self._kwparams.append(ts)

            elif p.annotation is bool:
                if v:
                    ts = f"{k.replace('_',' ')}: true"
                    self._kwparams.append(ts)

            elif p.annotation is List[str]:
                print("Treat as list of str")

        self._process_common_params(**kwargs)

        return None

    # let's parse and populate our default arguments according to
    # https://novem.no/docs/mail/sections/overview#common-parameters
    def _process_common_params(self, **kwargs: str) -> None:
        """
        Process the e-mail section parameters common to all novem mail
        #sections.
        * Padding
        * Margin
        * Borders
        * Colors

        We'll do some minor sanity checking here so we can report problems to
        the user, but ultimately the server side will verify the correctness
        of all parameters.
        """

        # valid direction list
        vdl = ["l", "r", "b", "t", "x", "y", "a"]
        valid_param_map = {
            "padding": ["multi", "dirsize"],
            "p": ["multi", "dirsize"],
            "margin": ["multi", "dirsize"],
            "m": ["multi", "dirsize"],
            "border": ["multi", "dirsize", "color"],
            "borders": ["multi", "dirsize", "color"],
            "b": ["multi", "dirsize", "color"],
            "foreground": ["color"],
            "fg": ["color"],
            "background": ["color"],
            "bg": ["color"],
        }

        vk = valid_param_map.keys()
        for k in kwargs.keys():

            # must be a common param
            if k not in vk:
                continue

            v = kwargs[k]

            # the content of the param must be a string or a list of strings
            if not (isinstance(v, str) or (isinstance(v, list) and not sum([1 for x in v if not isinstance(x, str)]))):

                # consider exception
                print((f'WARN: Section parameter "{k}" must be a string or ' "a list of strings"))

            # get some meta data about our param
            vp = valid_param_map[k]

            if "multi" not in vp and isinstance(v, list):
                # consider exception
                print(f'WARN: Section parameter "{k}" does not support lists')

            vs: List[str] = []
            # convert everything to a  list
            if isinstance(v, str):
                vs = [v]
            else:
                vs = v

            args = []
            # iterate over our v and construct our options
            for v in vs:
                if "dirsize" in vp:
                    splt = v.split(" ")
                    ds = splt[0]
                    dd = ds[0]  # direction
                    dz = 1
                    clr = ""
                    if dd not in vdl:
                        print(f'WARN: valid direction option for "{k}" ' f'is one of {",".join(vdl)}')
                    if len(ds) > 1:
                        try:
                            dz = int(ds[1])
                        except ValueError:
                            print(f'WARN: valid size option for "{k}" ' "is between 0 and 5")

                        if dz > 5 or dz < 0:
                            print(f'WARN: valid size option for "{k}" ' "is between 0 and 5")

                    if len(splt) > 1 and "color" in vp:
                        cls = splt[1:]
                        if len(cls) > 2:
                            print("WARN: a maximum of two colors can " f'be supplied for "{k}"')
                            cls = cls[:2]

                        # validate colors?
                        clr = f' {" ".join(cls)}'
                        # COLOR

                    args.append(f"{dd}{dz}{clr}")

                elif "color" in vp:
                    # only colors
                    cls = v.split(" ")
                    if len(cls) > 2:
                        print("WARN: a maximum of two colors can " f'be supplied for "{k}"')
                        cls = cls[:2]

                        # validate colors?
                    clr = f'{" ".join(cls)}'
                    args.append(f"{clr}")

            if len(args) == 1:
                self._cparams.append(f"{k}: {args[0]}")
            else:
                self._cparams.append(f'{k}: [{", ".join(args)}]')

    def get_markdown(self) -> str:
        """
        Returns novem e-mail markdown for this section
        """

        ot = f"{{{{ {self._type}"

        plist = self._params + self._kwparams + self._cparams

        # Add section controleld params first
        for p in plist:
            if not len(p):
                continue

            ot = f"{ot}\n  {p}"

        if len(plist):
            ot = f"{ot}\n"
        else:
            ot = f"{ot} "

        ot = f"{ot}}}}}"

        if self._subtype == "double":
            ot = f"{ot}\n{self._body}\n"
            ot = f"{ot}{{{{ /{self._type} }}}}"

        return ot

    def ispct(self, value: Optional[str]) -> Optional[str]:
        """
        Novem standard pct value check
        """
        if value:
            try:
                ww = int(str(value).replace("%", ""))
                value = f"{ww:d}%"
            except ValueError:
                value = "100%"

            return value

        return "100%"

    def isinval(self, value: Optional[str], default: Optional[str], opts: List[str]) -> Optional[str]:
        """
        Novem value in options check
        """
        if value:
            # verify alignment is in valid
            value = value.strip().lower()
            if value not in opts:
                # TODO: Raise exception for invalid value
                value = default
            return value

        return default


class VisSection(NovemEmailSectionApi):
    def __init__(
        self,
        vis: Plot,
        /,
        width: Optional[str] = "100%",
        align: Optional[str] = "center",
        include_caption: bool = True,
        include_title: bool = False,
        include_link: bool = False,
        override_title: str = None,
        override_caption: str = None,
        **kwargs: str,
    ) -> None:
        """Add a novem visual to your e-mail.

        :param vis: A handle to a :class:`novem.Plot` object referencing the
            visual you want to embed.
        :type vis: class:`novem.Plot`
        :param width: The width of the visual in the e-mail, defaults to 100%.
        :type width: str, optional
        :param align: Should the visual be aligned left, right or center.
        :type align: str, optional
        :param include_caption: Include the caption? Default true
        :type include_caption: bool, optional
        :param include_title: Include the title? Default false
        :type include_title: bool, optional
        :param include_link: Include a link to the web representation? Default
            false
        :type include_link: bool, optional
        :param override_caption: Override the caption with a string of your
            choice, leave out or set to None to ignore.
        :type override_caption: str, optional
        :param override_title: Override the title with a string of your
            choice, leave out or set to None to ignore.
        :type override_title: str, optional
        """
        # set our type
        self._type = "vis"
        self._subtype = "double"

        # verify our input variables, update them with default values
        width = self.ispct(width)
        align = self.isinval(align, "left", ["left", "right", "center"])

        # snapshot local variables
        locs = locals()

        # invoke super for param generations
        super().__init__(locs, signature(self.__class__.__init__).parameters, **kwargs)

        # add param for our mandatory section
        vsn = vis.shortname.strip()
        vsn = f"ref: {vsn}"
        self._params.append(vsn)


class CalloutSection(NovemEmailSectionApi):
    def __init__(
        self,
        markdown: str,
        /,
        type: Optional[str] = "info",
        desc: Optional[str] = None,
        cborder: str = None,
        **kwargs: str,
    ) -> None:
        """Add a novem visual to your e-mail.

        :param markdown: A markdown string that will be rendered inside the
            callout.
        :type vis: str

        :param type: The type of the callout, one of str:`warn`, str:`info`
            str:`err`, str:`ok`. Default is info
        :type type: str, optional

        :param desc: Short descriptive string that will show up before the
            callout.
        :type desc: str, optional

        :param cborder: Optional callout border type, one of str:`solid`,
            str:`dashed`. Set it to none and use the common border options for
            more control
        :type cborder: str, optional

        """
        # set our type
        self._type = "callout"
        self._subtype = "double"

        # verify our input variables, update them with default values
        type = self.isinval(
            type,
            "info",
            [
                "info",
                "warn",
                "err",
                "ok",
                "warning",
                "error",
                "alert",
                "success",
            ],
        )

        cborder = self.isinval(cborder, None, ["dashed", "solid"])

        # snapshot local variables
        locs = locals()

        # invoke super for param generations
        super().__init__(locs, signature(self.__class__.__init__).parameters, **kwargs)

        # add our body
        self._body = markdown


class PreviewSection(NovemEmailSectionApi):
    def __init__(
        self,
        markdown: str,
        /,
        **kwargs: str,
    ) -> None:
        """Add preview text to your mail.

        :param markdown: A markdown string that will be rendered to preview.
        :type vis: str
        """
        # set our type
        self._type = "preview"
        self._subtype = "double"

        # snapshot local variables
        locs = locals()

        # invoke super for param generations
        super().__init__(locs, signature(self.__class__.__init__).parameters, **kwargs)

        # add our body
        self._body = markdown


class ParagraphSection(NovemEmailSectionApi):
    def __init__(
        self,
        markdown: str,
        /,
        font_size: str = "m",
        font_style: str = "r",
        **kwargs: str,
    ) -> None:
        """Add preview text to your mail.

        :param markdown: A markdown string that will be rendered to preview.
        :type vis: str
        """
        # set our type
        self._type = "paragraph"
        self._subtype = "double"

        # snapshot local variables
        locs = locals()

        # invoke super for param generations
        super().__init__(locs, signature(self.__class__.__init__).parameters, **kwargs)

        # add our body
        self._body = markdown


class AuthorSection(NovemEmailSectionApi):
    def __init__(
        self,
        username: str,
        /,
        include_bio: bool = True,
        include_pciture: bool = True,
        override_bio: Optional[str] = None,
        **kwargs: str,
    ) -> None:
        """Add preview text to your mail.

        :param markdown: A markdown string that will be rendered to preview.
        :type vis: str
        """
        # set our type
        self._type = "author"
        self._subtype = "double"

        # snapshot local variables
        locs = locals()

        # invoke super for param generations
        super().__init__(locs, signature(self.__class__.__init__).parameters, **kwargs)

        # strip @s from username
        username = username.lower().replace("@", "")

        # add param for our mandatory section
        vsn = f"username: {username}"
        self._params.append(vsn)


class AttachmentSection(NovemEmailSectionApi):
    def __init__(
        self,
        ref: Plot,  # add grid and doc later
        /,
        format: str = "pdf",  # output format
        name: Optional[str] = None,  # optional name
        **kwargs: str,
    ) -> None:
        """Add a novem visual to your e-mail.

        :param ref: A handle to a :class:`novem.Plot` object referencing the
            visual you want to embed.
        :type ref: class:`novem.Plot`
        """
        # set our type
        self._type = "attachment"
        self._subtype = "double"

        # snapshot local variables
        locs = locals()

        # invoke super for param generations
        super().__init__(locs, signature(self.__class__.__init__).parameters, **kwargs)

        # add param for our mandatory section
        vsn = ref.shortname.strip()
        vsn = f"ref: {vsn}"
        self._params.append(vsn)


class NovemEmailSectionUtil(NovemEmailSection):
    def __init__(self) -> None:
        super().__init__()

    def get_markdown(self) -> str:
        """
        Returns novem e-mail markdown for this section
        """

        return self._body


class MarkdownSection(NovemEmailSectionUtil):
    def __init__(self, markdown: str) -> None:
        """Add markdown to your document

        :param markdown: A markdown string to add
        :type markdown: str
        """
        # invoke super for param generations
        super().__init__()

        # add our body
        self._body = markdown


class CodeSection(NovemEmailSectionUtil):
    def __init__(self, code: str, lang: str = "") -> None:
        """Add syntax highlighted code

        :param code: code text
        :type code: str
        :param lang: language
        :type lang: str
        """
        # invoke super for param generations
        super().__init__()

        # add our body
        self._body = f"""```{lang}\n{code}\n```"""


# TODO: Mail only section
# Code
# TODO: Mail only section
# Code



from __future__ import print_function

import os.path
import subprocess
import sys
import tempfile
from shutil import which
from typing import List, Optional

"""
This is utility code taken directly from https://github.com/fmoo/python-editor
on the 28th of February 2023. The below code is licensed Apache-2.0

Slight modification for integration with the novem cli
"""


__all__ = [
    "edit",
    "get_editor",
    "EditorError",
]

__version__ = "1.0.5"


class EditorError(RuntimeError):
    pass


def get_default_editors() -> List[str]:
    # TODO: Make platform-specific
    return [
        "editor",
        "vim",
        "emacs",
        "nano",
    ]


def get_editor_args(editor: str) -> List[str]:
    if editor in ["vim", "gvim", "vim.basic", "vim.tiny"]:
        return ["-f", "-o"]

    elif editor == "emacs":
        return ["-nw"]

    elif editor == "gedit":
        return ["-w", "--new-window"]

    elif editor == "nano":
        return ["-R"]

    elif editor == "code":
        return ["-w", "-n"]

    else:
        return []


def get_editor() -> str:
    # Get the editor from the environment.  Prefer VISUAL to EDITOR
    editor = os.environ.get("VISUAL") or os.environ.get("EDITOR")
    if editor:
        return editor

    # None found in the environment.  Fallback to platform-specific defaults.
    for ed in get_default_editors():
        path = which(ed)
        if path is not None:
            return path

    raise EditorError("Unable to find a viable editor on this system. Please consider setting your $EDITOR variable")


def get_tty_filename() -> str:
    if sys.platform == "win32":
        return "CON:"
    return "/dev/tty"


def edit(
    filename: Optional[str] = None,
    contents: Optional[str] = None,
    use_tty: Optional[bool] = None,
    suffix: str = "",
) -> str:
    editor = get_editor()
    args = [editor] + get_editor_args(os.path.basename(os.path.realpath(editor)))

    if use_tty is None:
        use_tty = sys.stdin.isatty() and not sys.stdout.isatty()

    if filename is None:
        tmp = tempfile.NamedTemporaryFile(suffix=suffix)
        filename = tmp.name

    if contents is not None:
        # For python3 only.  If str is passed instead of bytes, encode default
        wc = contents.encode()

        with open(filename, mode="wb") as f:
            f.write(wc)

    args += [filename]

    stdout = None
    if use_tty:
        stdout = open(get_tty_filename(), "wb")

    proc = subprocess.Popen(args, close_fds=True, stdout=stdout)
    proc.communicate()

    with open(filename, mode="rb") as f:
        return f.read().decode("utf-8")



from typing import Union

from novem.table import Selector

from ..api_ref import NovemAPI


class NovemColors(object):
    def __init__(self, api: NovemAPI) -> None:
        """ """
        self.api: NovemAPI = api
        super().__init__()

    def get(self) -> str:
        """
        Get list of all shares currently active
        """

        return self.api._read("/config/colors/colors")

    def set(self, colors: Union[str, Selector]) -> None:
        """
        replace all shares with the new set
        """

        cls: str = ""
        if isinstance(colors, str):
            cls = colors
        if isinstance(colors, Selector):
            cls = str(colors)
        else:
            cls = str(colors)

        return self.api._write("/config/colors/colors", cls)

    @property
    def type(self) -> str:
        """
        get the color type
        """
        return self.api._read("/config/colors/type")

    @type.setter
    def type(self, value: str) -> None:
        """
        Set the color type
        """
        return self.api._write("/config/colors/type", value)

    def __iadd__(self, style: Union[str, Selector]) -> str:
        """
        Add more styles to the plot
        """
        es = format(f"{self.get()}\n{str(style)}")
        return es

    def __repr__(self) -> str:
        return self.get()

    def __str__(self) -> str:
        return self.get()



class Applicator(object):
    """
    A novem base class for creating strings to apply to table
    selections
    """

    def __init__(self, string: str) -> None:
        self.string = string

    def __str__(self) -> str:
        return self.string



from typing import Final, Optional, Union

from novem.exceptions import NovemException

from ..applicator import Applicator

USEDATA: Final = "_"


# create color exception
class NovemColorException(NovemException):
    def __init__(self, message: str):

        message = f"Invalid color {message}"

        super().__init__(message)


class StaticColor(Applicator):
    def __init__(self, plane: str, color: str, dark: Optional[str] = None) -> None:

        # todo: validate colors and plane
        if plane not in ["bg", "fg"]:
            # TODO: Raise error
            plane = "bg"

        ref = f"{plane} {color}"

        if dark:
            ref = f"{ref} {dark}"

        # construct string
        super().__init__(ref)


class DynamicColor(Applicator):
    def __init__(
        self,
        plane: str,
        min: str = "bad",
        mid: Optional[str] = None,
        max: str = "good",
        vmin: Optional[Union[str, float, int]] = USEDATA,
        vmid: Optional[Union[str, float, int]] = USEDATA,
        vmax: Optional[Union[str, float, int]] = USEDATA,
        dmin: Optional[str] = None,
        dmid: Optional[str] = None,
        dmax: Optional[str] = None,
        interp: str = "lin",
    ) -> None:

        if plane not in ["bg", "fg"]:
            # TODO: Raise error
            plane = "bg"

        ccol = ""
        dcol = ""
        vr = ""

        nc = 3

        if not mid:
            nc = 2
            ccol = f"{min},{max}"
        else:
            ccol = f"{min},{mid},{max}"

        if nc == 2:
            vr = f"{vmin},{vmax}"
        else:
            vr = f"{vmin},{vmid},{vmax}"

        exp = interp
        if interp not in ["lin", "pow", "exp"]:
            exp = "lin"

        vr = f"({vr})^{exp}"

        if dmin:
            if nc == 2:
                dcol = f" {dmin},{dmax}{vr}"
            else:
                dcol = f" {dmin},{dmid},{dmax}{vr}"

        ref = f"{plane} {ccol}{vr}{dcol}"

        # construct string
        super().__init__(ref)



import getpass
import os
import random
import socket
import string
import sys
import urllib.request
from datetime import datetime
from typing import Any, Dict, Optional, Union

from novem.exceptions import Novem401

if os.name == "nt":
    from pyreadline3 import Readline  # type: ignore

    readline = Readline()
else:
    from signal import SIG_DFL, SIGPIPE, signal
    import readline  # type: ignore

from ..api_ref import NovemAPI
from ..utils import cl, colors, get_config_path, get_current_config
from ..version import __version__
from .common import mail, plot
from .config import check_if_profile_exists, update_config
from .group import group
from .invite import invite
from .setup import setup

sys.tracebacklimit = 0


def input_with_prefill(prompt: str, text: str) -> Any:
    def hook() -> Any:
        readline.insert_text(text)
        readline.redisplay()

    readline.set_pre_input_hook(hook)
    result = input(prompt)
    readline.set_pre_input_hook()
    return result


def do_update_config(
    profile: str,
    username: str,
    api_root: str,
    token_name: str,
    token: str,
    path: Optional[str],
) -> None:
    (status, path) = update_config(profile, username, api_root, token_name, token, path)

    print(f'{cl.OKGREEN} \u2713 {cl.ENDC}new token {cl.OKCYAN}"{token_name}"{cl.ENDC} created and saved to {path}')

    # save file


def refresh_config(args: Dict[str, Any]) -> None:
    (hasconf, curconf) = get_current_config(**args)

    if not hasconf:
        print("Configuration not found, please use --init to create it")

    token: Union[str, None] = None

    if "token" in args:
        token = args["token"]

    api_root: str = curconf["api_root"]

    assert "profile" in curconf
    profile: str = curconf["profile"]

    # let's grab our token
    ignore_ssl = False
    if "ignore_ssl_warn" in curconf:
        ignore_ssl = curconf["ignore_ssl_warn"]

    valid_char_sm = string.ascii_lowercase + string.digits
    valid_char = valid_char_sm + "-_"
    hostname: str = socket.gethostname()

    token_name: Union[str, None] = None
    if not token_name:
        token_hostname: str = "".join([x for x in hostname.lower() if x in valid_char])
        nounce: str = "".join(random.choice(valid_char_sm) for _ in range(8))
        token_name = f"np-{token_hostname}-{nounce}".lower()[-32:]

    new_token_name = "".join([x for x in token_name if x in valid_char])

    if token_name != new_token_name:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            "The supplied token name contained invalid charracters,"
            f' token changed to "{cl.OKCYAN}{new_token_name}{cl.ENDC}"'
        )
        token_name = new_token_name

    print(f"Refresh token for profile: {profile}")

    # get novem username
    assert "username" in curconf
    prefill = curconf["username"]

    username = input_with_prefill(" \u2022 novem username: ", prefill)
    # username = "abc"

    # get novem password
    password = getpass.getpass(" \u2022 novem password: ")

    # authenticate and request token by name
    req = {
        "username": username,
        "password": password,
        "token_name": token_name,
        "token_description": (f'cli token created for "{hostname}" ' f'on "{datetime.now():%Y-%m-%d:%H:%M:%S}"'),
    }

    novem = NovemAPI(api_root=api_root, ignore_config=True, ignore_ssl=ignore_ssl)

    try:
        res = novem.create_token(req)
        token = res["token"]
        token_name = res["token_name"]

    except Novem401:
        print("Invalid username and/or password")
        sys.exit(1)

    # update token
    do_update_config(profile, username, api_root, token_name, token, None)


def init_config(args: Dict[str, Any]) -> None:
    """
    Initialize user and config

    the --init flag has been supplied, this means we are going to update the
    configuration file. There are several different scenarios we have to adapt
    to in this case:

    * a simple --init with no options and missing config file
      * request username and password from user
      * authenticate against service
      * create config file and store credentials if successful

    * a simple --init with no options and existing config file
      * check if config has a default user
      * if default user exist in config, inform user that config
      * already exist and terminate

    * a simple --init with --profile <username> and missing config file
      * request username and password from user
        * with username prefilled as --profile
      * authenticate against service
      * create config file and store credentials if successful

      --init

      --profile
      --token
      --api-url
      --force

      user in config

      config exist



      if --token not supplied, request username and password
      if --profile specified, prefill username

      if token supplied request token info, if username password

    """

    if not args:
        # make mypy happy, return if no argument supplied
        return

    _do_debug = False
    if "debug" in args and args["debug"]:
        _do_debug = True

    token: Union[str, None] = None

    if "token" in args:
        token = args["token"]

    api_root: str = args["api-url"]
    profile: str = args["profile"]
    force: str = args["force"]
    config_path: str = args["config_path"]

    # first check if we have a valid config
    if _do_debug:
        print("INIT: check if profile exists")
    profile_exists: bool = check_if_profile_exists(profile, config_path)
    if profile_exists and not force:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            f' The supplied profile "{cl.OKCYAN}{profile}{cl.ENDC}" already '
            f"exist, use --force to override"
        )
        sys.exit(1)

    if _do_debug:
        print("INIT: construct token name")
    valid_char_sm = string.ascii_lowercase + string.digits
    valid_char = valid_char_sm + "-_"
    hostname: str = socket.gethostname()

    token_name: Union[str, None] = None
    if not token_name:
        token_hostname: str = "".join([x for x in hostname.lower() if x in valid_char])
        nounce: str = "".join(random.choice(valid_char_sm) for _ in range(8))
        token_name = f"np-{token_hostname}-{nounce}".lower()[-32:]

    new_token_name = "".join([x for x in token_name if x in valid_char])

    if token_name != new_token_name:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            "The supplied token name contained invalid charracters,"
            f' token changed to "{cl.OKCYAN}{new_token_name}{cl.ENDC}"'
        )
        token_name = new_token_name

    # get novem username
    prefill = ""
    if "profile" in args:
        prefill = args["profile"]

    if _do_debug:
        print("INIT: request user input")

    username = input_with_prefill(" \u2022 novem username: ", prefill)
    # username = "abc"

    # get novem password
    password = getpass.getpass(" \u2022 novem password: ")

    # authenticate and request token by name
    req = {
        "username": username,
        "password": password,
        "token_name": token_name,
        "token_description": (f'cli token created for "{hostname}" ' f'on "{datetime.now():%Y-%m-%d:%H:%M:%S}"'),
    }

    if _do_debug:
        print("INIT: construct request")

    if not api_root:
        _, curconf = get_current_config(**args)
        api_root = curconf["api_root"]

    # let's grab our token
    ignore_ssl = args.get("ignore_ssl", False)

    if _do_debug:
        print("INIT: request token")

    novem = NovemAPI(api_root=api_root, ignore_config=True, ignore_ssl=ignore_ssl)

    try:
        res = novem.create_token(req)
        token = res["token"]
        token_name = res["token_name"]

    except Novem401:
        print("Invalid username and/or password")
        sys.exit(1)

    # default profile to username if not supplied
    if not profile:
        profile = username

    # let's write our config
    do_update_config(profile, username, api_root, token_name, token, config_path)


def print_short(parser: Any) -> None:
    parser.print_usage()

    print()
    print("Novem command line interface - no options supplied")
    print()
    print("  novem -h, --help      print a concise help")
    print("  novem --guide         print the comprehensive novem cli guide")
    print()
    print("  novem --init          if you already have an account and want to get started")
    print()
    print("  novem -p              list your plots")
    print("  novem -m              list your mails")


def run_cli_wrapped() -> None:
    colors()

    raw_args = sys.argv[1:]

    #    return

    # Gather the provided arguements as an array.
    # (parser:Any, args:Dict[str, str]) = setup(raw_args)
    (parser, args) = setup(raw_args)

    # let's add traceback back in
    if "debug" in args and args["debug"]:
        del sys.tracebacklimit

    if len(raw_args) == 0:
        print_short(parser)
        return

    if args and args["version"]:
        print(f"novem {__version__}")
        return

    if args and "debug" in args and args["debug"]:
        """
        Print some sytem information
        """
        import platform
        import socket

        # TODO: use argument supplied config path instead

        print("System information")
        print()
        print(f"Platform: {platform.platform()}")
        pv = sys.version.split("\n")
        print(f"Python: {pv}")
        print()
        proxy = urllib.request.getproxies()
        print(f"Proxy settings: {proxy}")
        hname = socket.gethostname()
        print(f"Hostname: {hname}")
        api_ip = socket.gethostbyname("api.novem.no")
        print(f"API IP : {api_ip}")
        main_ip = socket.gethostbyname("novem.no")
        print(f"Core IP: {main_ip}")
        print()
        conf_dir, conf_path = get_config_path()
        print(f"Current config folder: {conf_dir}")
        print(f"Folder exists: {os.path.exists(conf_dir)}")
        print(f"Folder is writeable: {os.access(conf_dir, os.W_OK)}")

        print()
        print(f"Current config path:  {conf_path}")
        print(f"File exists: {os.path.isfile(conf_path)}")
        print(f"File is writeable: {os.access(conf_path, os.W_OK)}")

        # TODO: add dns and gateway

        print()

    # we are getting an init instruction
    if args and args["init"]:
        init_config(args)
        return

    # verify profile
    if args and args["profile"]:
        config_path: str = args["config_path"]
        profile_exists: bool = check_if_profile_exists(args["profile"], config_path)
        if not profile_exists:
            print(
                f"""\
Profile "{args["profile"]}" doesn't exist in your config. Please add it using:
novem --init --profile {args["profile"]}\
"""
            )

            sys.exit(1)

    # we are getting a refresh instruction
    if args and args["refresh"]:
        refresh_config(args)
        return

    # check info and if present get info
    if args and args["info"]:
        novem = NovemAPI(**args)
        info = novem.read("/admin/profile/overview")
        print(info)
        return

    # if --fs is set get terminal dimensions and ammend qpr
    if args and "fs" in args and args["fs"]:
        sz = os.get_terminal_size()
        qpr = ""
        if "qpr" in args and args["qpr"]:
            qpr = f"{args['qpr']},"

        qpr = f"{qpr}cols={sz.columns},rows={sz.lines-1}"
        args["qpr"] = qpr

    # operate on plot
    if args and args["plot"] != "":
        plot(args)
    elif args and args["mail"] != "":
        mail(args)
    elif args and args["invite"] != "":
        invite(args)
    elif args and ("group" in args or "org" in args):
        group(args)
        pass


def run_cli() -> None:
    if os.name != "nt":
        signal(SIGPIPE, SIG_DFL)  # supress broken pipe error
    try:
        run_cli_wrapped()
    except KeyboardInterrupt:
        pass


__all__ = ["run_cli"]



import os
import sys
import urllib.request
from typing import Any, Dict, List, Optional, Tuple

import requests

from novem.exceptions import Novem403, Novem404

from ..api_ref import NovemAPI
from ..utils import cl
from ..utils import colors as clrs
from ..version import __version__
from .files import NovemFiles
from .shared import NovemShare

s = requests.Session()


class NovemVisAPI(NovemAPI):
    """ """

    _hdr_post: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
        "Content-type": "text/plain",
    }
    _hdr_put: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
    }
    _hdr_get: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
    }
    _hdr_del: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
    }

    _proxies: Dict[Any, Any] = {}

    shared: Optional[NovemShare] = None
    files: Optional[NovemFiles] = None

    _vispath: Optional[str] = None
    _debug: bool = False

    def __init__(self, **kwargs: Any) -> None:

        self._proxies = urllib.request.getproxies()

        super().__init__(**kwargs)

        self.user = None

        if self._config["ignore_ssl_warn"] or ("ignore_ssl" in kwargs and kwargs["ignore_ssl"]):
            # supress ssl warnings
            s.verify = False
            import urllib3

            urllib3.disable_warnings()

        if "debug" in kwargs and kwargs["debug"]:
            self._debug = True

        if "create" not in kwargs or kwargs["create"]:
            # let's create our plot if -C specified, always
            # create when used as an api unless specifically told not to
            self.api_create("")

        if "user" in kwargs and kwargs["user"]:
            self.user = kwargs["user"]

        if "qpr" in kwargs and kwargs["qpr"]:
            self._qpr = kwargs["qpr"].replace(",", "&")

        self.shared = NovemShare(self)
        self.files = NovemFiles(self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "shared" and self.shared:
            self.shared.set(value)
        else:
            super().__setattr__(name, value)

    def api_dump(self, outpath: str) -> None:
        """
        Iterate over current id and dump output to supplied path
        """

        qpath = f"{self._api_root}vis/{self._vispath}/{self.id}/"

        if self.user:
            qpath = f"{self._api_root}users/{self.user}/vis/" f"{self._vispath}/{self.id}/"

        # create util function
        def rec_tree(path: str) -> None:
            qp = f"{qpath}{path}"
            fp = f"{outpath}{path}"
            # print(f"QP: {qp}")
            req = s.get(
                qp,
                auth=("", self.token),
                headers=self._hdr_get,
                proxies=self._proxies,
            )

            if not req.ok:
                return None

            headers = req.headers
            try:
                tp = headers["X-NS-Type"]
            except KeyError:
                tp = "file"

            # if i am a file, write to disc
            if tp == "file":
                with open(fp, "w") as f:
                    f.write(req.text)
                print(f"Writing file:    {fp}")
                return None

            # if I am a folder, make a folder and recurse
            os.makedirs(fp, exist_ok=True)
            print(f"Creating folder: {fp}")

            nodes: List[Dict[str, str]] = req.json()

            # Recurse relevant structure
            for r in [x for x in nodes if x["type"] not in ["system_file", "system_dir"]]:
                rec_tree(f'{path}/{r["name"]}')

        # start recurison
        rec_tree("")

    def api_tree(self, colors: bool = False, relpath: str = "/") -> str:
        """
        Iterate over the current id and print a "pretty" ascii tree
        """
        if relpath[0] != "/":
            relpath = f"/{relpath}"

        clrs()

        qpath = f"{self._api_root}vis/{self._vispath}/{self.id}{relpath}"

        if self.user:
            qpath = f"{self._api_root}users/{self.user}/vis/" f"{self._vispath}/{self.id}{relpath}"

        # TODO: we're using some hard coded unicode symbols and colors here
        # probably better to make this configurable by the user and perhaps
        # even move it to cli

        # this is an ugly hack, and mostly here to help users familiarize
        # themselves with the api structure

        # some display options
        c = "├"
        b = "└"
        v = "│"
        h = "─"

        # create util function
        def rec_tree(path: str, level: int = 0, last: List[bool] = [False]) -> Tuple[List[str], str]:
            qp = f"{qpath}{path}"
            req = s.get(
                qp,
                auth=("", self.token),
                headers=self._hdr_get,
                proxies=self._proxies,
            )

            if not req.ok:
                return ([], "")

            headers = req.headers

            try:
                tp = headers["X-NS-Type"]
            except KeyError:
                tp = "file"

            if tp == "file":
                print("The tree display is only available for `dir` paths")
                sys.exit(-1)

            nodes: List[Dict[str, str]] = req.json()

            hdp = []
            if level == 0:
                try:
                    hdp = headers["X-NS-Permissions"].split(", ")
                except KeyError:
                    hdp = []

            pfx = ""
            for il in last:
                if il:
                    pfx += "    "
                else:
                    pfx += f"{v}   "

            # drop system stuff
            nodes = [x for x in nodes if x["type"] not in ["system_file", "system_dir"]]

            resp = ""
            # convert element into a tree structure
            nodes = sorted(nodes, key=lambda k: (k["type"], k["name"]))
            for r in nodes:
                rd = "r" if "r" in r["permissions"] else "-"
                w = "w" if "w" in r["permissions"] else "-"
                d = "d" if "d" in r["permissions"] else "-"
                if colors:
                    a = f"{cl.FGGRAY}[{rd}{w}{d}]{cl.ENDC}"
                else:
                    a = f"[{rd}{w}{d}]"

                if r["name"] == nodes[-1]["name"]:
                    mc = last + [True]
                    co = f"{b}"
                else:
                    mc = last + [False]
                    co = f"{c}"

                if r["type"] == "dir":
                    if colors:
                        resp += f"{pfx}{co}{h}{h} {a} {cl.OKBLUE}" f'{r["name"]}/{cl.ENDC}\n'
                    else:
                        resp += f'{pfx}{co}{h}{h} {a} {r["name"]}/\n'

                    resp += rec_tree(f'{path}/{r["name"]}', level + 1, mc)[1]
                else:
                    resp += f'{pfx}{co}{h}{h} {a} {r["name"]}\n'

            # order by dir, files, alphabetically
            return (hdp, resp)

        hdp, tr = rec_tree("/", 0, [True])

        sf = f"{self.id}{relpath}"
        if sf[-1] != "/":
            sf = f"{sf}/"

        if colors:
            sf = f"{cl.OKBLUE}{sf}{cl.ENDC}"

        rd = "r" if "r" in hdp else "-"
        w = "w" if "w" in hdp else "-"
        d = "d" if "d" in hdp else "-"
        if colors:
            a = f"{cl.FGGRAY}[{rd}{w}{d}]{cl.ENDC}"
        else:
            a = f"[{rd}{w}{d}]"
        tr = f"{a} {sf}\n{tr}"

        return tr[:-1]  # strip trailing newline

    def api_read(self, relpath: str) -> str:
        """
        Read the api value located at realtive path
        """

        qpath = f"{self._api_root}vis/{self._vispath}/{self.id}{relpath}"

        # We can read information from other users, but not perform any
        # other actions so only the GET method supports the custom user
        # pathing
        if self.user:
            qpath = f"{self._api_root}users/{self.user}/vis/" f"{self._vispath}/{self.id}{relpath}"

        if self._qpr and len(self._qpr):
            qpath = f"{qpath}?{self._qpr}"

        if self._debug:
            print(f"GET: {qpath}")

        r = s.get(
            qpath,
            auth=("", self.token),
            headers=self._hdr_get,
            proxies=self._proxies,
        )

        # TODO: verify result and raise exception if not ok
        if r.status_code == 404:
            raise Novem404(qpath)

        if r.status_code == 403:
            raise Novem403

        return r.content.decode("utf-8")

    def api_read_bytes(self, relpath: str) -> bytes:
        qpath = f"{self._api_root}vis/{self._vispath}/{self.id}{relpath}"

        # We can read information from other users, but not perform any
        # other actions so only the GET method supports the custom user
        # pathing
        if self.user:
            qpath = f"{self._api_root}users/{self.user}/vis/" f"{self._vispath}/{self.id}{relpath}"

        if self._qpr and len(self._qpr):
            qpath = f"{qpath}?{self._qpr}"

        if self._debug:
            print(f"GET: {qpath}")

        r = s.get(
            qpath,
            auth=("", self.token),
            headers=self._hdr_get,
            proxies=self._proxies,
        )

        # TODO: verify result and raise exception if not ok
        if r.status_code == 404:
            raise Novem404(qpath)

        if r.status_code == 403:
            raise Novem403

        return r.content

    def api_delete(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        if self.user:
            print(f"you cannot modify another users {self._vispath}")
            return

        path = f"{self._api_root}vis/{self._vispath}/{self.id}{relpath}"

        if self._debug:
            print(f"DELETE: {path}")

        r = s.delete(
            path,
            auth=("", self.token),
            headers=self._hdr_del,
            proxies=self._proxies,
        )

        if r.status_code == 404:
            raise Novem404(path)

        if r.status_code == 403:
            raise Novem403

        # TODO: verify result and raise exception if not ok
        if not r.ok:
            print(r)
            print(f"DELETE: {path}")
            print("body")
            print("---")
            print(r.text)
            print("headers")
            print("---")
            print(r.headers)
            print("should raise an error")

    def api_create(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        if self.user:
            print(f"you cannot modify another users {self._vispath}")
            return

        path = f"{self._api_root}vis/{self._vispath}/{self.id}{relpath}"

        if self._debug:
            print(f"PUT: {path}")

        r = s.put(
            path,
            auth=("", self.token),
            headers=self._hdr_put,
            proxies=self._proxies,
        )

        if r.status_code == 404:
            raise Novem404(path)

        if r.status_code == 403:
            raise Novem403(path)

        if r.status_code == 409:
            # we will ignore 409 errors
            # as creating objects that already exist is not a problem
            return

        # TODO: verify result and raise exception if not ok
        if not r.ok:
            print(r)
            print(f"PUT: {path}")
            print("body")
            print("---")
            print(r.text)
            print("headers")
            print("---")
            print(r.headers)
            print("should raise a general error")

    def api_write(self, relpath: str, value: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        if self.user:
            print(f"you cannot modify another users {self._vispath}")
            return

        path = f"{self._api_root}vis/{self._vispath}/{self.id}{relpath}"

        if self._debug:
            print(f"POST: {path}")

        r = s.post(
            path,
            auth=("", self.token),
            headers=self._hdr_post,
            data=value.encode("utf-8"),
            proxies=self._proxies,
        )

        if r.status_code == 404:
            raise Novem404(path)

        if r.status_code == 403:
            raise Novem403

        # TODO: verify result and raise exception if not ok
        if not r.ok:
            print(r)
            print(f"POST: {path} {value}")
            print("body")
            print("---")
            print(r.text)
            print(r.status_code)
            print("headers")
            print("---")
            for k, v in r.headers.items():
                print(f"   {k}: {v}")
            print("should raise a general error")

    @property
    def qpr(self) -> str:
        if self._qpr:
            return self._qpr.replace("&", ",")
        else:
            return ""

    @qpr.setter
    def qpr(self, value: str) -> None:
        if value:
            self._qpr = value.replace(",", "&")



import json
from typing import List, Union

from novem.exceptions import Novem404

from ..api_ref import NovemAPI


class NovemShare(object):
    """
    Novem share
    """

    # api: NovemAPI = None

    def __init__(self, api: NovemAPI) -> None:
        """ """
        self.api: NovemAPI = api

    def get(self) -> List[str]:
        """
        Get list of all shares currently active
        """

        try:
            s = self.api.api_read("/shared")
            shares = json.loads(s)
            shared = [x["name"] for x in shares]

        except Novem404:
            shared = []

        return shared

    def set(self, share: Union[str, List[str]]) -> None:
        """
        replace all shares with the new set
        """

        if isinstance(share, str):
            shares = [share]
        else:
            shares = share

        es = self.get()

        rms = set(es) - set(shares)
        adds = set(shares) - set(es)

        for r in rms:
            self.api.api_delete(f"/shared/{r}")

        for a in adds:
            self.api.api_create(f"/shared/{a}")

    def __iadd__(self, share: str) -> List[str]:
        """
        Add a new share to the plot
        """
        es = self.get()
        es.append(share)

        return es

    def __isub__(self, share: str) -> List[str]:
        """
        Remove a share from the plot
        """
        es = self.get()
        ns = set(es) - set([share])

        return list(ns)



from __future__ import annotations

from typing import Optional, Union

from novem.table import Selector

from ..api_ref import NovemAPI

# from novem.table import CellAlign, CellBorder, CellFormat,
# CellPadding, CellText, CellWidth, CellMerge


class IProxy(str):

    p: Optional[NovemCellConfig] = None
    path: str = ""

    # TODO figure out how we can return a new string on demand
    # def __init__(self, parent: Any, path: str) -> None:
    #    self.p = parent
    #    self.path = path

    def __iadd__(self, val: Union[str, Selector]) -> str:
        if not self.p:
            return ""
        cur = self.p.read(self.path)
        es = format(f"{cur}\n{str(val)}")
        return es

    def __str__(self) -> str:
        if not self.p:
            return ""
        return self.p.read(self.path)

    def __repr__(self) -> str:
        if not self.p:
            return ""
        return self.p.read(self.path)


class NovemCellConfig(object):
    def __init__(self, api: NovemAPI) -> None:
        """ """
        self.api: NovemAPI = api
        super().__init__()

    def read(self, path: str) -> str:
        """
        Get list of all shares currently active
        """
        return self.api._read(path)

    def write(self, path: str, value: Union[str, Selector]) -> None:
        """
        replace all shares with the new set
        """

        vls: str = ""
        if isinstance(value, str):
            vls = value
        if isinstance(value, Selector):
            vls = str(value)
        else:
            vls = str(value)

        return self.api._write(path, vls)

    def _proxy(self, path: str) -> str:
        ip = IProxy()
        ip.p = self
        ip.path = path
        return ip

    @property
    def align(self) -> str:
        return self._proxy("/config/table/cell/align")

    @align.setter
    def align(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/align", style)

    @property
    def border(self) -> str:
        return self._proxy("/config/table/cell/border")

    @border.setter
    def border(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/border", style)

    @property
    def format(self) -> str:
        return self._proxy("/config/table/cell/format")

    @format.setter
    def format(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/format", style)

    @property
    def padding(self) -> str:
        return self._proxy("/config/table/cell/padding")

    @padding.setter
    def padding(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/padding", style)

    @property
    def text(self) -> str:
        return self._proxy("/config/table/cell/text")

    @text.setter
    def text(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/text", style)

    @property
    def width(self) -> str:
        return self._proxy("/config/table/cell/width")

    @width.setter
    def width(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/width", style)

    @property
    def merge(self) -> str:
        return self._proxy("/config/table/cell/merge")

    @merge.setter
    def merge(self, style: Union[str, Selector]) -> None:
        return self.write("/config/table/cell/merge", style)



from typing import Optional, TypedDict

from typing_extensions import NotRequired

Config = TypedDict(
    "Config",
    {
        "username": NotRequired[str],
        "token": Optional[str],
        "api_root": str,
        "ignore_ssl_warn": bool,
        "profile": NotRequired[str],
    },
)



from __future__ import annotations

from typing import Any, List, Optional, Union

from novem.exceptions import Novem404
from novem.vis import NovemVisAPI

from .mail_sections import NovemEmailSection, PreviewSection


class Mail(NovemVisAPI):
    """
    Novem mail class
    """

    def __init__(self, id: str, **kwargs: Any) -> None:
        """
        :id mail name, duplicate entry will update the mail

        :to  TO recipients
        :cc  CC recipients
        :bcc BCC recipients
        :subject subject line
        """

        # if we have an @ name we will override id and user
        if id[0] == "@":
            cand = id[1:].split("~")
            id = cand[1]
            uname = cand[0]
            kwargs["user"] = uname

        self.id = id

        self._vispath = "mails"
        self._type = "mail"

        super().__init__(**kwargs)

        self._section_updated: bool = False

        self._preview: Optional[NovemEmailSection] = None

        self._sections: List[NovemEmailSection] = []

        self._parse_kwargs(**kwargs)

    def __call__(self, content: Any, **kwargs: Any) -> Any:
        """
        Set's the content of the mail

        The parameter either needs to be a text string
        of CSV formatted text or an object with a to_csv
        function.
        """

        raw_str: str = str(content)

        # setting the content object will invoke a server write
        self.content = raw_str

        # also update our varibales
        self._parse_kwargs(**kwargs)

        # return the original object so users can chain the contentframe
        return content

    def _parse_kwargs(self, **kwargs: Any) -> None:

        # first let our super do it's thing
        super()._parse_kwargs(**kwargs)

        # get a list of valid properties
        # exclude content as it needs to be run last
        props = [x for x in dir(self) if x[0] != "_" and x not in ["content", "read", "delete", "write", "status"]]

        do_content = False
        do_status = False
        for k, v in kwargs.items():
            if k == "content":
                do_content = True

            if k == "status":
                do_status = True

            if k not in props:
                continue

            # print(f"{k} :: {v}")
            # set our value
            setattr(self, k, v)

        if do_content:
            content = kwargs["content"]
            setattr(self, "content", content)

        # status can send e-mail, so it's the very last thing to be updated
        if do_status:
            status = kwargs["status"]
            setattr(self, "status", status)

    # we'll implement generic properties common across all plots here
    def _produce_content(self) -> str:
        """
        Produce a valid novem e-mail markdown file
        from the supplied options
        """

        # construct our markdown
        ctnt = ""

        if self._preview:
            ctnt = self._preview.get_markdown()

        if len(self._sections):
            ctnt += "\n\n" + "\n\n".join([s.get_markdown() for s in self._sections])

        return ctnt

    def add_section(self, section: NovemEmailSection) -> Mail:
        """
        Add section to e-mail

        :section NovemEmailSection to add
        """

        if not isinstance(section, NovemEmailSection):
            # throw?
            print("You must supply a valid NovemEmailSection")

        if isinstance(section, PreviewSection):
            # update our preview section
            self._preview = section
            self._section_updated = True
        else:
            # append the section to our section list
            self._sections.append(section)
            self._section_updated = True

        return self
        # verify that

    def render(self) -> None:
        """
        Write e-mail to server
        """
        if self._section_updated:
            ctnt = self._produce_content()
            self.api_write("/content", ctnt)
            # update after api_write in case of exception
            self._section_updated = False

        return None

    def send(self) -> None:
        """
        Send e-mail to recipients
        """
        # check if there are any recipients registered
        try:
            to = self.to
        except Novem404:
            to = ""

        try:
            cc = self.cc
        except Novem404:
            cc = ""

        try:
            bcc = self.bcc
        except Novem404:
            bcc = ""

        reps = f"{to}\n{cc}\n{bcc}".split("\n")
        reps = [x for x in reps if x != ""]

        if len(reps) == 0:
            # print("No recipients registered, e-mail won't be sent")
            return None

        self.render()
        return self.api_write("/status", "sending")

    def _send(self) -> None:
        return self.api_write("/status", "sending")

    def test(self) -> None:
        """
        Create a test e-mail
        """
        # no need to check for recipients,

        self.render()
        return self.api_write("/status", "testing")

    @property
    def name(self) -> str:
        return self.api_read("/name").strip()

    @name.setter
    def name(self, value: str) -> None:
        return self.api_write("/name", value)

    @property
    def description(self) -> str:
        return self.api_read("/description")

    @description.setter
    def description(self, value: str) -> None:
        return self.api_write("/description", value)

    @property
    def summary(self) -> str:
        return self.api_read("/summary")

    @summary.setter
    def summary(self, value: str) -> None:
        return self.api_write("/summary", value)

    @property
    def url(self) -> str:
        return self.api_read("/url").strip()

    @property
    def shortname(self) -> str:
        return self.api_read("/shortname").strip()

    @property
    def log(self) -> str:
        return self.api_read("/log").strip()

    ###
    # Important attributes that cause render and sending
    ###

    @property
    def content(self) -> str:
        return self.api_read("/content")

    @content.setter
    def content(self, value: str) -> None:
        return self.api_write("/content", value)

    @property
    def status(self) -> str:
        return self.api_read("/status")

    @status.setter
    def status(self, value: str) -> None:
        return self.api_write("/status", value)

    ###
    # Recipients
    ###

    # recipients to string
    def _r2s(self, rcpts: List[str]) -> str:
        return "\n".join(rcpts)

    # recipients to list
    def _r2l(self, rcpts: str) -> List[str]:
        return rcpts.split("\n")

    @property
    def to(self) -> str:
        return self.api_read("/recipients/to")

    @to.setter
    def to(self, value: Union[str, List[str]]) -> None:
        if not value:
            return
        if isinstance(value, list):
            value = self._r2s(value)
        elif isinstance(value, str):
            value = "\n".join(value.split(","))
        return self.api_write("/recipients/to", value)

    @property
    def cc(self) -> str:
        return self.api_read("/recipients/cc")

    @cc.setter
    def cc(self, value: Union[str, List[str]]) -> None:
        if not value:
            return
        if isinstance(value, list):
            value = self._r2s(value)
        elif isinstance(value, str):
            value = "\n".join(value.split(","))
        return self.api_write("/recipients/cc", value)

    @property
    def bcc(self) -> str:
        return self.api_read("/recipients/bcc")

    @bcc.setter
    def bcc(self, value: Union[str, List[str]]) -> None:
        if not value:
            return
        if isinstance(value, list):
            value = self._r2s(value)
        elif isinstance(value, str):
            value = "\n".join(value.split(","))
        return self.api_write("/recipients/bcc", value)

    ###
    # Config
    ###

    @property
    def subject(self) -> str:
        return self.api_read("/config/subject")

    @subject.setter
    def subject(self, value: str) -> None:
        if not value:
            return
        return self.api_write("/config/subject", value)

    @property
    def theme(self) -> str:
        return self.api_read("/config/theme")

    @theme.setter
    def theme(self, value: str) -> None:
        return self.api_write("/config/theme", value)

    @property
    def size(self) -> str:
        return self.api_read("/config/size")

    @size.setter
    def size(self, value: str) -> None:
        return self.api_write("/config/size", value)

    @property
    def template(self) -> str:
        return self.api_read("/config/template")

    @template.setter
    def template(self, value: str) -> None:
        return self.api_write("/config/template", value)

    @property
    def reply_to(self) -> str:
        return self.api_read("/config/reply_to")

    @reply_to.setter
    def reply_to(self, value: str) -> None:
        return self.api_write("/config/reply_to", value)



import configparser
import os
import stat
from os import path
from pathlib import Path
from typing import Optional, Tuple

from ..utils import get_config_path


def update_config(
    profile: str,
    username: str,
    api_root: str,
    token_name: str,
    token: str,
    path: Optional[str],
) -> Tuple[bool, str]:
    """
    Write configuration to file
    """

    novem_config: str = ""
    if not path:
        (novem_dir, novem_config) = get_config_path()

        # create path and file if not exist
        Path(novem_dir).mkdir(parents=True, exist_ok=True)
        Path(novem_config).touch(mode=stat.S_IRUSR | stat.S_IWUSR, exist_ok=True)
        os.chmod(novem_config, stat.S_IRUSR | stat.S_IWUSR)  # ensure file is not world readable
    else:
        novem_config = path

    config = configparser.ConfigParser()

    # read our config object
    config.read(novem_config)

    add_api_root = True

    # check if config has general section
    if not config.has_section("general"):
        config.add_section("general")
        config.set("general", "profile", profile)
        config.set("general", "api_root", api_root)
        add_api_root = False
    else:
        gar = config["general"]["api_root"]
        if gar == api_root:
            # don't add api_root to user config if it's the same as general api
            # root. If the user want's to modify the global api_url then this
            # let's them not have to update all users.
            # the user can still add individual api_roots if they want when
            # creating users

            add_api_root = False
        if "profile" not in config["general"]:
            # incomplete config, set our user to be default
            config.set("general", "profile", username)

    # TODO: Expand default app configs here when needed
    if not config.has_section("app:cli"):
        config.add_section("app:cli")

    if not config.has_section("app:pylib"):
        config.add_section("app:pylib")

    if not config.has_section("app:fuse"):
        config.add_section("app:fuse")

    profile_name = f"profile:{profile}"

    # add/update our section
    if not config.has_section(profile_name):
        config.add_section(profile_name)

    config.set(profile_name, "username", username)
    if add_api_root:
        config.set(profile_name, "api_root", api_root)

    config.set(profile_name, "token_name", token_name)
    config.set(profile_name, "token", token)

    with open(novem_config, "w+") as configfile:
        config.write(configfile)

    return (True, novem_config)


def check_if_profile_exists(profile: str, config_path: str) -> bool:
    """
    Check if config dir already contains a valid token for the profile
    """

    if not config_path:
        (novem_dir, novem_config) = get_config_path()
    else:
        novem_config = config_path

    # check if path exists
    if not path.exists(novem_config):
        # profile don't exist

        return False

    config = configparser.ConfigParser()

    # read our config object
    config.read(novem_config)

    profile_name = f"profile:{profile}"

    # add/update our section
    if config.has_section(profile_name):
        return True

    return False



import datetime
import email.utils as eut
import json
from typing import Any, Dict, List

from novem.exceptions import Novem404

from ..api_ref import NovemAPI
from ..utils import cl, pretty_format


def list_invites(args: Dict[str, Any], novem: NovemAPI) -> None:
    """
    List pending invites
    """

    # see if list flag is set

    ilist = []

    try:
        ilist = json.loads(novem.read("/admin/invites/"))
    except Novem404:
        ilist = []

    ilist = sorted(ilist, key=lambda x: x["name"])

    if args["list"]:

        # print to terminal
        for p in ilist:
            print(p["name"])

        return

    flist = []

    for i in ilist:
        res = {}
        nm = i["name"]
        # let's populate our final list with info
        res["name"] = nm
        res["id"] = nm

        gt = "unkown"
        user = ""
        group = ""
        org = ""
        if nm[0] == "+" and "~" in nm:
            gt = "organisation group"
            spl = nm[1:].split("~")
            org = spl[0]
            group = spl[1]
        elif nm[0] == "@" and "~" in nm:
            gt = "user group"
            spl = nm[1:].split("~")
            user = spl[0]
            group = spl[1]
        elif nm[0] == "+" and "~" not in nm:
            gt = "organisation"
            org = nm[1:]
        else:
            gt = "unkown"

        res["group"] = group
        if org:
            res["org_user"] = f"+{org}"
        else:
            res["org_user"] = f"@{user}"

        res["type"] = gt
        res["created"] = i["created_on"]

        flist.append(res)

    ppo: List[Dict[str, Any]] = [
        {
            "key": "id",
            "header": "Invitation ID",
            "type": "text",
            "overflow": "keep",
        },
        # {
        #    "key": "id",
        #    "header": "ID",
        #    "type": "text",
        #    "overflow": "keep",
        # },
        {
            "key": "type",
            "header": "Type",
            "type": "text",
            "clr": cl.OKCYAN,
            "overflow": "keep",
        },
        {
            "key": "group",
            "header": "Group",
            "type": "text",
            "overflow": "truncate",
        },
        {
            "key": "org_user",
            "header": "Org / User",
            "type": "text",
            "overflow": "truncate",
        },
        {
            "key": "created",
            "header": "Created",
            "type": "date",
            "overflow": "keep",
        },
    ]

    for p in flist:
        nd = datetime.datetime(*eut.parsedate(p["created"])[:6])
        p["created"] = nd.strftime("%Y-%m-%d %H:%M")

    ppl = pretty_format(flist, ppo)

    print(ppl)


def invite(args: Dict[str, Any]) -> None:

    novem = NovemAPI(**args)

    # we are invoked so plot must exist
    invite_name = args["invite"]

    if invite_name is None:
        # we need to list plots
        list_invites(args, novem)
        return

    # check if
    if "accept" in args and args["accept"]:
        novem.write(f"/admin/invites/{invite_name}/accept", "yes")

    elif "reject" in args and args["reject"]:
        novem.write(f"/admin/invites/{invite_name}/accept", "no")



import os
import sys
from typing import Any, Dict, Literal, Optional

from novem import Mail, Plot
from novem.api_ref import Novem404, NovemAPI
from novem.cli.editor import edit
from novem.cli.setup import Share
from novem.cli.vis import list_vis, list_vis_shares
from novem.utils import data_on_stdin
from novem.vis import NovemVisAPI


class VisBase:
    def __init__(self, type: Literal["mail", "plot"]) -> None:
        self.type = type

    @property
    def title(self) -> str:
        return self.type.capitalize()

    @property
    def fragment(self) -> str:
        return f"{self.type}s"

    def set_data(self, nva: NovemVisAPI, data: str) -> None:
        if self.type == "mail":
            nva.content = data
        else:
            nva.data = data

    def mk(self, name: str, user: Optional[str], ignore_ssl: bool, create: bool, args: Dict[str, Any]) -> NovemVisAPI:
        if self.type == "mail":
            return Mail(
                name,
                user=user,
                ignore_ssl=ignore_ssl,
                create=create,
                config_path=args["config_path"],
                to=args["to"],
                cc=args["cc"],
                bcc=args["bcc"],
                subject=args["subject"],
                qpr=args["qpr"],
                debug=args["debug"],
                profile=args["profile"],
            )

        elif self.type == "plot":
            return Plot(
                name,
                user=user,
                ignore_ssl=ignore_ssl,
                create=create,
                config_path=args["config_path"],
                qpr=args["qpr"],
                debug=args["debug"],
                profile=args["profile"],
            )
        else:
            assert False, "Invalid type"

    def __call__(self, args: Dict[str, Any]) -> None:
        # we are invoked so vis must exist
        name = args[self.type]

        if name is None:
            # we need to list plots
            list_vis(args, self.title)
            return

        # if delete flag is set, we need to delete it
        if args["delete"]:
            # creating a plot just to delete it seems wasteful
            # We'll just use the raw api
            novem = NovemAPI(**args)

            try:
                novem.delete(f"vis/{self.fragment}/{name}")
                return
            except Novem404:
                print(f"{self.title} {name} did not exist")
                sys.exit(1)

        usr = None
        if "for_user" in args and args["for_user"]:
            usr = args["for_user"]

        # we are going to modify one of our plots, listing and deletion has
        # already been handled so we can use the plot api without wasting
        # "resources'

        # this will create the plot if it does not exist
        ignore_ssl = False
        if "ignore_ssl" in args:
            ignore_ssl = args["ignore_ssl"]

        create = args["create"]

        vis = self.mk(name=name, user=usr, ignore_ssl=ignore_ssl, create=create, args=args)

        # this is a data dump instruction, we'll ignore everything else
        if "dump" in args and args["dump"]:
            path = args["dump"]

            print(f'Dumping api tree structure to "{path}"')
            vis.api_dump(outpath=path)
            return

        # if we detect a tree query then we'll discard all other IO
        if "tree" in args and args["tree"] != -1:
            path = args["tree"]
            if not path:
                path = "/"

            ts = vis.api_tree(colors=True, relpath=path)
            print(ts)
            return

        # if we have the -e or edit flag then this takes presedence over all other
        # inputs
        if "edit" in args and args["edit"]:
            path = args["edit"]

            # fetch our target and warn if it doesn't exist
            ctnt = vis.api_read(f"/{path}")

            # get new content
            nctnt = edit(contents=ctnt, use_tty=True)

            if ctnt != nctnt:
                # update content
                vis.api_write(f"/{path}", nctnt)

        else:

            ptype = args["type"]
            # print(ptype)
            if ptype:
                vis.type = ptype

            found_stdin = False
            stdin_data = data_on_stdin()
            stdin_has_data = bool(stdin_data)

            # check if we have any explicit inputs [-w's]
            if args["input"] and len(args["input"]):
                # we have inputs
                for i in args["input"]:
                    path = f"/{i[0]}"

                    if len(i) == 1:
                        if stdin_has_data and not found_stdin:
                            assert stdin_data
                            ctnt = stdin_data

                            vis.api_write(path, ctnt)
                            found_stdin = True
                        elif found_stdin:
                            print("stdin can only be sent to a single destination per invocation")
                            sys.exit(1)
                        else:
                            print(f'No data found on stdin, "-w {path}" requires data to be supplied on stdin')
                            sys.exit(1)

                    elif len(i) == 2 and i[1][0] == "@":
                        fn = os.path.expanduser(i[1][1:])
                        try:
                            with open(fn, "r") as f:
                                ctnt = f.read()
                                vis.api_write(path, ctnt)
                        except FileNotFoundError:
                            print(f'The supplied input file "{fn}" does not exist. Please review your options')
                            sys.exit(1)

                    else:
                        ctnt = i[1]
                        vis.api_write(path, ctnt)

            # check if we have standard input and not already "used" it
            if not found_stdin and stdin_has_data:
                assert stdin_data
                ctnt = stdin_data
                # got stdin data
                self.set_data(vis, ctnt)

        # check if we are changing any permissions
        share_op, share_target = args["share"]
        if share_op is Share.CREATE:
            # add a share to the vis
            vis.shared += share_target  # type: ignore

        if share_op is Share.DELETE:
            # remove a share from the vis
            vis.shared -= share_target  # type: ignore

        if share_op is Share.LIST:
            # check if we should print our shares, we will not provide other outputs
            list_vis_shares(name, args, self.title)
            return

        # Output - we only allow a singular -o or -x and will return as soon as
        # we find one

        # -x takes presedence
        x = args["tc"]
        if x:
            # get file endpoint
            if os.name == "nt":
                from colorama import just_fix_windows_console  # type: ignore

                just_fix_windows_console()
            print(vis.files.ansi, end="")  # type: ignore

            # exit as we only support one output
            return

        # TODO: check if we are reading any valus from the plot [-r or -x]
        out = args["out"]
        if out:
            outp = vis.api_read(f"/{out}")
            print(outp, end="")


def mail(args: Dict[str, Any]) -> None:
    mail = VisBase("mail")
    mail(args)


def plot(args: Dict[str, Any]) -> None:
    plot = VisBase("plot")
    plot(args)



import datetime
import email.utils as eut
import json
import re
from typing import Any, Dict, List

from novem.exceptions import Novem404

from ..api_ref import NovemAPI
from ..utils import cl, pretty_format


def list_orgs(args: Dict[str, Any], novem: NovemAPI, path: str) -> None:
    ilist = []

    try:
        ilist = json.loads(novem.read(path))
    except Novem404:
        ilist = []

    ilist = sorted(ilist, key=lambda x: x["name"])

    if "filter" in args and args["filter"]:
        fv = args["filter"]
        if fv[0] != "^":
            fv = f".*{fv}"
        if fv[-1] != "$":
            fv = f"{fv}.*"

        flt = re.compile(fv, re.I)

        ilist = [x for x in ilist if (flt.match(x["name"]))]

    flist = []

    if args["list"]:

        # print to terminal
        for p in ilist:
            print(p["name"])

        return

    for i in ilist:
        res = {}
        nm = i["name"]
        # let's populate our final list with info
        res["name"] = nm
        res["id"] = nm

        gt = "unkown"
        user = ""
        group = ""
        org = ""
        gt = "organisation"

        res["group"] = group
        if org:
            res["org_user"] = f"+{org}"
        else:
            res["org_user"] = f"@{user}"

        res["type"] = gt
        res["created"] = i["created_on"]

        flist.append(res)

    ppo: List[Dict[str, Any]] = [
        {
            "key": "id",
            "header": "Group ID",
            "type": "text",
            "overflow": "keep",
        },
        # {
        #    "key": "id",
        #    "header": "ID",
        #    "type": "text",
        #    "overflow": "keep",
        # },
        {
            "key": "type",
            "header": "Type",
            "type": "text",
            "clr": cl.OKCYAN,
            "overflow": "keep",
        },
        # {
        #    "key": "group",
        #    "header": "Group",
        #    "type": "text",
        #    "overflow": "truncate",
        # },
        # {
        #    "key": "org_user",
        #    "header": "Org / User",
        #    "type": "text",
        #    "overflow": "truncate",
        # },
        {
            "key": "created",
            "header": "Created",
            "type": "date",
            "overflow": "keep",
        },
    ]

    for p in flist:
        nd = datetime.datetime(*eut.parsedate(p["created"])[:6])
        p["created"] = nd.strftime("%Y-%m-%d %H:%M")

    ppl = pretty_format(flist, ppo)

    print(ppl)


def list_groups(args: Dict[str, Any], novem: NovemAPI, path: str) -> None:
    """
    List pending invites
    """

    # see if list flag is set

    ilist = []

    try:
        ilist = json.loads(novem.read(path))
    except Novem404:
        ilist = []

    ilist = sorted(ilist, key=lambda x: x["name"])

    if "filter" in args and args["filter"]:
        fv = args["filter"]
        if fv[0] != "^":
            fv = f".*{fv}"
        if fv[-1] != "$":
            fv = f"{fv}.*"

        flt = re.compile(fv, re.I)

        ilist = [x for x in ilist if (flt.match(x["name"]))]

    if args["list"]:

        # print to terminal
        for p in ilist:
            print(p["name"])

        return

    flist = []

    orgo = ""
    if "org" in args and args["org"]:
        orgo = args["org"]

    for i in ilist:
        res = {}
        nm = i["name"]
        # let's populate our final list with info
        res["name"] = nm
        res["id"] = nm

        gt = "unkown"
        user = ""
        group = ""
        org = ""
        if nm[0] == "+" and "~" in nm:
            gt = "organisation group"
            spl = nm[1:].split("~")
            org = spl[0]
            group = spl[1]
        elif nm[0] == "@" and "~" in nm:
            gt = "user group"
            spl = nm[1:].split("~")
            user = spl[0]
            group = spl[1]
        elif nm[0] == "+" and "~" not in nm:
            gt = "organisation"
            org = nm[1:]
        else:
            gt = "unkown"

        res["group"] = group
        if org:
            res["org_user"] = f"+{org}"
        else:
            res["org_user"] = f"@{user}"

        res["type"] = gt

        res["created"] = i["created_on"]

        if orgo:
            res["type"] = "organisation group"
            res["org_user"] = f"+{orgo}"
            res["group"] = nm
            res["id"] = f"+{orgo}~{nm}"

        flist.append(res)

    ppo: List[Dict[str, Any]] = [
        {
            "key": "id",
            "header": "Group ID",
            "type": "text",
            "overflow": "keep",
        },
        # {
        #    "key": "id",
        #    "header": "ID",
        #    "type": "text",
        #    "overflow": "keep",
        # },
        {
            "key": "type",
            "header": "Type",
            "type": "text",
            "clr": cl.OKCYAN,
            "overflow": "keep",
        },
        {
            "key": "group",
            "header": "Group",
            "type": "text",
            "overflow": "truncate",
        },
        {
            "key": "org_user",
            "header": "Org / User",
            "type": "text",
            "overflow": "truncate",
        },
        {
            "key": "created",
            "header": "Created",
            "type": "date",
            "overflow": "keep",
        },
    ]

    for p in flist:
        nd = datetime.datetime(*eut.parsedate(p["created"])[:6])
        p["created"] = nd.strftime("%Y-%m-%d %H:%M")

    ppl = pretty_format(flist, ppo)

    print(ppl)


# TODO: shift this logic to a group class
# and make it available from python as well
def group(args: Dict[str, Any]) -> None:
    novem = NovemAPI(**args)

    has_org = "org" in args and args["org"] != ""
    has_group = "group" in args and args["group"] != ""
    has_user = "for_user" in args and args["for_user"] != ""

    # we are invoked so plot must exist
    group_name = args["group"] if "group" in args else None
    org_name = args["org"] if "org" in args else None
    # user_name = args["for_user"] if "for_user" in args else None

    # print(f"{has_group} - {has_org} - {has_user}")
    # print(f"{group_name} - {org_name} - {user_name}")

    # check if we should create a group
    if has_group and group_name and args["create"] and not has_org:
        # create usergroup
        path = f"/admin/groups/{group_name}/"
        novem.create(path)
        return

    # check if we should create a group
    if has_group and group_name and args["create"] and has_org and org_name:
        # create usergroup
        path = f"/admin/orgs/{org_name}/groups/{group_name}/"
        novem.create(path)
        return

    # check if we should create a group
    if has_group and group_name and args["delete"] and not has_org:
        # create usergroup
        path = f"/admin/groups/{group_name}/"
        novem.delete(path)
        return

    # check if we should create a group
    if has_group and group_name and args["delete"] and has_org and org_name:
        # create usergroup
        path = f"/admin/orgs/{org_name}/groups/{group_name}/"
        novem.delete(path)
        return

    # check if we should invite a user
    if "invite_user" in args and args["invite_user"] and has_group and has_org:
        iu = args["invite_user"]
        path = f"/admin/orgs/{org_name}/groups/{group_name}/roles/members/{iu}"
        novem.create(path)
        return

    if "invite_user" in args and args["invite_user"] and has_group and group_name and not has_org:
        iu = args["invite_user"]
        path = f"/admin/groups/{group_name}/roles/members/{iu}"
        novem.create(path)
        return

    if has_group and has_org and not has_user and org_name and not group_name:
        list_groups(args, novem, f'/orgs/{args["org"]}/groups/')
        return

    if not has_group and not has_user and has_org and not org_name:
        list_orgs(args, novem, "/orgs/")
        return

    if has_group and not has_user and not has_org and not group_name:
        list_groups(args, novem, "/groups/")
        return

    # print("no condition")



from typing import Any, List, Optional

import numpy as np
import pandas as pd

from novem.exceptions import NovemException


class NovemSelectorException(NovemException):
    def __init__(self, message: str):

        message = f"Invalid selector format {message}."

        super().__init__(message)


def enhance_positions(positions: List[int], ior: int) -> List[int]:
    if not positions or not isinstance(ior, int):
        return positions
    if ior == 0:
        return positions
    elif ior < 0:
        return list(range(min(positions) + ior, min(positions))) + positions
    else:  # ior > 0
        return positions + list(range(max(positions) + 1, max(positions) + 1 + ior))


class Selector(object):
    """
    The novem Selector is a convenience function for carving out parts of a
    dataframe for styling

    The purpose of a novem selector is to supply it with a sliced dataframe
    along with the original and the selector will figure out the correct novem
    selector syntax to replicate the sliced dataframe.

    However, due to some idiosyncracies of pandas data frames it's important
    to know the following:

    Novem selectors doesn't have an index or column concept, all rows and
    columns supplied to the api is considered when applying styling, this
    is different from pandas where index and columns are not part of the
    slicing result.

    This means that by default loc and iloc results are offset by 1 in the
    novem output relative to what you would expect from a pandas dataframe

    Similarly, a .loc[:,:] or .iloc[:,:] will produce 1: 1: not : :

    To override this behavior you have the c (column) and i (index) options
    you can manually specify to override the derived values

    There also cannot be duplicate row or index values when using the slicer

    If you have duplicate values we recommend you reset_index and supply
    a co or ci of -1
    """

    def __init__(
        self,
        selector: Any,
        applicator: Any,
        r: Any = None,
        c: Optional[str] = None,
        i: Optional[str] = None,
        co: Optional[int] = None,
        io: Optional[int] = None,
        cor: Optional[int] = None,
        ior: Optional[int] = None,
    ) -> None:
        """ """

        # TODO: verify that it's a valid novem selector before
        # storing it and throw an error if not
        self.selector = selector
        self.applicator = applicator
        self.ref = r
        self.c = c
        self.i = i
        self.co = co
        self.io = io
        self.cor = cor
        self.ior = ior

    def _pd_ix_lookup(self) -> str:

        filter = self.selector
        frame = self.ref

        if frame is None:
            # TODO: Raise exception
            raise NovemSelectorException("We need a reference to correctly infer pandas selections")

        if filter is None:
            raise NovemSelectorException("There is no data in our dataframe, we need at " "least one value.")

        # Grab levels
        cl = filter.index.nlevels
        rl = 1

        co = 0
        io = 0
        if self.co:
            co = self.co
        if self.io:
            io = self.io

        cor = 0
        ior = 0
        if self.cor:
            cor = self.cor
        if self.ior:
            ior = self.ior

        # Check if the filter is a DataFrame or a Series
        if isinstance(filter, pd.DataFrame):
            row_indices = filter.index
            col_indices = filter.columns
            rl = filter.columns.nlevels
        else:
            row_indices = filter.index
            col_indices = pd.Index([filter.name])

            # a series can indicate both a single column
            # and a single slicer, we'll have to check key membership
            # on our original frame to be sure, this is not perfect
            # as if all the values are in both the columns and index we
            # have to just guess
            if all(value in frame.columns for value in row_indices):
                row_indices = pd.Index([filter.name])
                col_indices = filter.index
                rl = filter.index.nlevels
                cl = 1

        # TODO: We need to check against values in the frame for our
        # filtered dataset, else we are still failing to account for
        # duplicate values outside of filter

        # Find the row and column positions in the original dataframe
        row_positions = []
        used_positions = set()
        for row_label in row_indices:
            position = frame.index.get_loc(row_label)
            if isinstance(position, np.ndarray):
                position = np.flatnonzero(position)
                position = [pos for pos in position if pos not in used_positions]
                used_positions.update(position[:1])
                row_positions.append(position[0])
            else:
                if position not in used_positions:
                    used_positions.add(position)
                    row_positions.append(position)

        col_positions = []
        used_col_positions = set()
        for col_label in col_indices:
            position = frame.columns.get_loc(col_label)
            if isinstance(position, np.ndarray):
                position = np.flatnonzero(position)
                position = [pos for pos in position if pos not in used_col_positions]
                used_col_positions.update(position[:1])
                col_positions.append(position[0])
            else:
                if position not in used_col_positions:
                    used_col_positions.add(position)
                    col_positions.append(position)

        # convert to novem 0 based index and drop negative offset results
        height = len(frame)
        width = len(frame.columns)
        row_positions = [x + rl + io for x in row_positions if x < height]
        row_positions = enhance_positions(row_positions, ior)
        row_positions = [x for x in row_positions if x >= 0 and x <= height]

        col_positions = [x + cl + co for x in col_positions if x < width]
        col_positions = enhance_positions(col_positions, cor)
        col_positions = [x for x in col_positions if x >= 0 and x <= width]

        # Create a comma-separated list of row and column positions
        row_str = ",".join(map(str, row_positions))
        col_str = ",".join(map(str, col_positions))

        if self.i:
            row_str = self.i

        if self.c:
            col_str = self.c

        # Combine the row and column positions with the provided text
        result = f"{row_str} {col_str}"
        return result

    def get_selector_string(self) -> str:
        """ """
        if isinstance(self.selector, str):
            return self.selector

        return self._pd_ix_lookup()

    def __str__(self) -> str:
        """
        return a string representaion of our selector
        """

        return f"{self.get_selector_string()} {self.applicator}"



import argparse as ap
import shutil
from enum import Enum
from typing import Any, Dict, Tuple

width = min(120, shutil.get_terminal_size().columns - 2)


class Share(Enum):
    NOT_GIVEN = 0
    CREATE = 1
    DELETE = 2
    LIST = 3


def formatter(prog: str) -> ap.RawDescriptionHelpFormatter:
    return ap.RawDescriptionHelpFormatter(prog, width=width)


def setup(raw_args: Any = None) -> Tuple[Any, Dict[str, str]]:
    parser = ap.ArgumentParser(
        prog="novem",
        description="Novem commandline interface.",
        formatter_class=formatter,
    )

    parser.add_argument(
        "--ignore-ssl",
        dest="ignore_ssl",
        action="store_true",
        required=False,
        default=False,
        help=ap.SUPPRESS,
    )

    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        required=False,
        default=False,
        help=ap.SUPPRESS,
    )

    parser.add_argument(
        "--dump",
        metavar=("OUT_PATH"),
        dest="dump",
        action="store",
        required=False,
        default=None,
        help=ap.SUPPRESS,
    )

    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        required=False,
        default=False,
        help="print the current version and terminate",
    )

    parser.add_argument(
        "--color",
        dest="color",
        action="store_true",
        required=False,
        default=False,
        help="preserve colors when redirecting to pipe or file",
    )

    parser.add_argument(
        "--api-url",
        dest="api-url",
        action="store",
        required=False,
        default=None,
        help="api url to use, defaul is https://api.novem.no/v1/",
    )

    parser.add_argument(
        "--config-path",
        "-c",
        dest="config_path",
        action="store",
        required=False,
        default=None,
        help="specify configuration file to use",
    )

    parser.add_argument(
        "--profile",
        dest="profile",
        action="store",
        required=False,
        default=None,
        help="which user to use, combine with --init to setup a new profile and --force to override an existing one",
    )

    parser.add_argument(
        "--token",
        dest="token",
        action="store",
        required=False,
        default=None,
        help="use this token instead, overrides profile lookup",
    )

    parser.add_argument(
        "--info",
        dest="info",
        action="store_true",
        required=False,
        help="print info about the current user",
    )

    setup = parser.add_argument_group("setup")

    setup.add_argument(
        "--init",
        dest="init",
        action="store_true",
        help="authenticate with the novem service and create default configuration",
    )

    setup.add_argument(
        "--force",
        dest="force",
        action="store_true",
        required=False,
        help="force reinit of existing profile",
    )

    setup.add_argument(
        "--token-name",
        dest="token-name",
        action="store",
        required=False,
        default=None,
        help="name of token (lowercase alphanumeric, no whitespace)",
    )

    parser.add_argument(
        "--refresh",
        dest="refresh",
        action="store_true",
        required=False,
        default=False,
        help="refresh the token for the current profile (or the one supplied by --profile), "
        "requires username and password",
    )

    vis = parser.add_argument_group("common visualisation arguments")

    vis.add_argument(
        "-C",
        dest="create",
        action="store_true",
        required=False,
        help="create the visualisation if it doesn't exist",
    )

    vis.add_argument(
        "-D",
        dest="delete",
        action="store_true",
        help="delete the current visualisation defined by -[pdmgv] or share defined by -s",
    )

    vis.add_argument(
        "-s",
        dest="share",
        action="store",
        required=False,
        default="",
        nargs="?",
        help="select a share group to operate on, no parameter will list all current shares",
    )

    vis.add_argument(
        "-l",
        dest="list",
        action="store_true",
        help="print ids only, no pretty printing",
    )

    vis.add_argument(
        "-f",
        metavar=("REGEX"),
        required=False,
        dest="filter",
        action="store",
        help="filter visualisations by rgex",
    )

    # support multiple inputs
    vis.add_argument(
        "-w",
        dest="input",
        action="append",
        nargs="+",
        metavar=("PATH", "VALUE"),
        help="write the supplied VALUE to the given PATH. PATH is mandatory. VALUE can be an explicit value, "
        "a filename prefixed with @ or data on stdin",
    )

    vis.add_argument(
        "-r",
        dest="out",
        action="store",
        required=False,
        default=None,
        metavar=("PATH"),
        help="read the content of PATH and prints it to stdout",
    )

    vis.add_argument(
        "-e",
        dest="edit",
        action="store",
        required=False,
        default=None,
        metavar=("PATH"),
        help="open the content located at PATH in $EDITOR and update the saved content on editor exit",
    )

    vis.add_argument(
        "-u",
        metavar=("USER"),
        dest="for_user",
        default=ap.SUPPRESS,
        action="store",
        required=False,
        help="specify user to view shared visualisation from",
    )

    vis.add_argument(
        "-o",
        metavar=("SOURCE"),
        dest="for_other",
        action="store",
        required=False,
        default=None,
        help="specify entity to view vis for, @username, +org, @username~usergroup or +org~orggroup are supported",
    )

    vis.add_argument(
        "--tree",
        metavar=("PATH"),
        dest="tree",
        action="store",
        required=False,
        default=-1,
        nargs="?",
        help="print a tree overview of the api structure at the given path, all input/output options are ignored",
    )

    term = parser.add_argument_group("terminal")

    term.add_argument(
        "-x",
        dest="tc",
        action="store_true",
        required=False,
        default=False,
        help="shorthand for requesting a terminal friendly output, identical to doing -r files/plot.ansi",
    )

    term.add_argument(
        "--qpr",
        dest="qpr",
        action="store",
        required=False,
        default=None,
        help="comma separated list of query parameters to include with request such as "
        "cols=$COLUMNS,rows=$(($lines-1))",
    )

    term.add_argument(
        "--fs",
        dest="fs",
        action="store_true",
        required=False,
        default=False,
        help='shorthand for creating a "full screen" version of the terminal vis',
    )

    # Currently not added as it would expand on our dependencies
    # we might consider adding it or providing it as a separate package
    # in the future
    if 0:
        term.add_argument(
            "--watch",
            dest="watch",
            action="store_true",
            required=False,
            default=False,
            help="connect to the server and redraws the visual when new information is available",
        )

    plot = parser.add_argument_group("plot")

    plot.add_argument(
        "-p",
        dest="plot",
        action="store",
        required=False,
        default="",
        nargs="?",
        help="select plot to operate on, no parameter will list all your plots",
    )

    plot.add_argument(
        "-t",
        dest="type",
        action="store",
        required=False,
        default=None,
        help="shorthand for setting the type of the plot, identical to doing -w config/type TYPE",
    )

    mail = parser.add_argument_group("mail")

    mail.add_argument(
        "-m",
        dest="mail",
        action="store",
        required=False,
        default="",
        nargs="?",
        help="select mail to operate on, no parameter will list all your mails",
    )

    mail.add_argument(
        "--to",
        dest="to",
        metavar=("RECIPIENTS"),
        action="store",
        required=False,
        default=None,
        help="shorthand for setting recipient of mail, identical to doing -w recipients/to RECIPIENTS",
    )

    mail.add_argument(
        "--cc",
        dest="cc",
        metavar=("RECIPIENTS"),
        action="store",
        required=False,
        default=None,
        help="shorthand for setting recipient of mail, identical to doing -w recipients/cc RECIPIENTS",
    )

    mail.add_argument(
        "--bcc",
        dest="bcc",
        metavar=("RECIPIENTS"),
        action="store",
        required=False,
        default=None,
        help="shorthand for setting recipient of mail, identical to doing -w recipients/bcc RECIPIENTS",
    )

    mail.add_argument(
        "--subject",
        dest="subject",
        metavar=("SUBJECT"),
        action="store",
        required=False,
        default=None,
        help="shorthand for setting subject of mail, identical to doing -w config/subject SUBJECT",
    )

    mail.add_argument(
        "-S",
        dest="send",
        action="store_true",
        required=False,
        help="send the e-mail to recipients",
    )

    mail.add_argument(
        "-T",
        dest="test",
        action="store_true",
        required=False,
        help="send a test e-mail to your registered address",
    )

    invite = parser.add_argument_group("invite")

    invite.add_argument(
        "-i",
        dest="invite",
        action="store",
        required=False,
        default="",
        nargs="?",
        help="select invite to operate on, no parameter will list all pending invitations",
    )

    invite.add_argument(
        "--accept",
        dest="accept",
        action="store_true",
        required=False,
        default=False,
        help="accept the invite",
    )

    invite.add_argument(
        "--reject",
        dest="reject",
        action="store_true",
        required=False,
        default=False,
        help="reject the invite",
    )

    group = parser.add_argument_group(
        "group",
        description="""\
Operate on novem groups.

-C, --create - create
-D, --delete - delete the group
--invite      - invite a member to a group
--remove      - manage members

Examples:
  --invite bob -C analysts
""",
    )

    group.add_argument(
        "-O",
        dest="org",
        action="store",
        required=False,
        default=ap.SUPPRESS,
        nargs="?",
        help="select an organisation operate on, no parameter will list all organisations of which you are a member",
    )

    group.add_argument(
        "-G",
        dest="group",
        action="store",
        required=False,
        default=ap.SUPPRESS,
        nargs="?",
        help="""\
select an organisation -O or user -u group operate on.
No parameter will list all organisations groups of which you are a member""",
    )

    group.add_argument(
        "--invite",
        metavar=("USER"),
        dest="invite_user",
        action="store",
        required=False,
        help="invite a USER to the current organisation",
    )

    # group.add_argument(
    #    "--role",
    #    dest="role",
    #    action="store",
    #    required=False,
    #    help="specify role to give invited user, empty means member"
    # )

    args = vars(parser.parse_args(raw_args))

    # fix up the --share option
    share = args.pop("share")
    if share == "":
        args["share"] = (Share.NOT_GIVEN, None)
    elif share is None:
        args["share"] = (Share.LIST, None)
    elif args["create"]:
        args["create"] = None
        args["share"] = (Share.CREATE, share)
    elif args["delete"]:
        args["delete"] = None
        args["share"] = (Share.DELETE, share)
    else:
        args["share"] = None

    return (parser, args)



from novem.exceptions import Novem404

from ..api_ref import NovemAPI


class NovemFiles(object):
    """
    Novem Files
    """

    # api: NovemAPI = None

    def __init__(self, api: NovemAPI) -> None:
        """ """
        self.api: NovemAPI = api

    def get(self, fn: str) -> str:
        try:
            ctnt = self.api.api_read(f"/files/{fn}")
        except Novem404:
            ctnt = ""

        return ctnt

    @property
    def txt(self) -> str:
        """
        return novem text file
        """
        return self.get(f"{self.api._type}.txt")

    @property
    def ansi(self) -> str:
        """
        return novem ansi representation file
        """
        return self.get(f"{self.api._type}.ansi")

    @property
    def img(self) -> bytes:
        """
        return novem png representation file
        """
        return self.api.api_read_bytes(f"/files/{self.api._type}.png")



import sys
import urllib.request
from typing import Any, Dict, Optional

import requests

from .utils import get_current_config

s = requests.Session()


class NovemException(Exception):
    pass


class Novem404(NovemException):
    def __init__(self, message: str):

        # 404 errors can occur if users are not authenticated, let them know
        # future improvement: consider requesting a fixed endpoint (like
        # whoami) and notify if not authenticated
        message = f"Resource not found: {message} (Are you authenticated?)"

        super().__init__(message)


class Novem403(NovemException):
    pass


class Novem401(NovemException):
    pass


class NovemAPI(object):
    """
    Novem API class

    * Read config file
    * Communicate with api.novem.no
    * Offer utilities for subclasses
    """

    id: Optional[str] = None
    _type: Optional[str] = None
    _qpr: Optional[str] = None

    def __init__(self, **kwargs: Any) -> None:
        """ """

        config_status, config = get_current_config(**kwargs)

        self._config = config

        if self._config["ignore_ssl_warn"]:
            # supress ssl warnings
            s.verify = False
            import urllib3

            urllib3.disable_warnings()

        # api root should always be supplied in the result
        self._api_root = config["api_root"]

        if config.get("token", None):
            assert config["token"]
            self.token = config["token"]

        elif not config_status:
            print(
                """\
Novem config file is missing.  Either specify config file location with
the config_path parameter, or setup a new token using
$ python -m novem --init
"""
            )
            sys.exit(0)

        if self._api_root[-1] != "/":
            # our code assumes that the api_root ends with a /
            self._api_root = f"{self._api_root}/"

    def _parse_kwargs(self, **kwargs: Any) -> None:
        """
        Parse the arguments and invoke the novem api
        """

        if "api_root" in kwargs:
            self._api_root = kwargs["api_root"]

    def create_token(self, params: Dict[str, str]) -> Dict[str, str]:

        r = s.post(
            f"{self._api_root}token",
            json=params,
        )

        if not r.ok:
            resp = r.json()
            if r.status_code == 401:
                raise Novem401(resp["message"])

        res = r.json()

        return res

    def delete(self, path: str) -> bool:

        r = s.delete(
            f"{self._api_root}{path}",
            auth=("", self.token),
        )

        if not r.ok:
            resp = r.json()
            if r.status_code == 404:
                raise Novem404(resp["message"])
            else:
                print(r.json())

        return r.ok

    def read(self, path: str) -> str:

        r = s.get(
            f"{self._api_root}{path}",
            auth=("", self.token),
        )

        if not r.ok:
            resp = r.json()
            if r.status_code == 404:
                raise Novem404(resp["message"])

        return r.text

    def write(self, path: str, value: str) -> None:

        r = s.post(
            f"{self._api_root}{path}",
            auth=("", self.token),
            headers={
                "Content-type": "text/plain",
            },
            data=value.encode("utf-8"),
            proxies=urllib.request.getproxies(),
        )

        if not r.ok:
            resp = r.json()
            if r.status_code == 404:
                raise Novem404(resp["message"])
            else:
                print(r.json())

    def create(self, path: str) -> None:

        r = s.put(
            f"{self._api_root}{path}",
            auth=("", self.token),
            proxies=urllib.request.getproxies(),
        )

        if not r.ok:
            resp = r.json()
            if r.status_code == 404:
                raise Novem404(resp["message"])
            else:
                print(r.json())

    def _read(self, relpath: str) -> str:
        """
        Read the api value located at realtive path
        """
        ...

    def _write(self, relpath: str, value: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass

    def api_read(self, relpath: str) -> str:
        """
        Read the api value located at realtive path
        """
        ...

    def api_read_bytes(self, relpath: str) -> bytes:
        """
        Read the api value located at realtive path
        """
        ...

    def api_delete(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass

    def api_create(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass

    def api_write(self, relpath: str, value: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass



from typing import Any, Dict, Optional

from novem.vis import NovemVisAPI

from .cell import NovemCellConfig
from .colors import NovemColors
from .plot_config import NovemPlotConfig


class Plot(NovemVisAPI):
    """
    Novem plot class
    """

    colors: Optional[NovemColors] = None
    cell: Optional[NovemCellConfig] = None
    config: Optional[NovemPlotConfig] = None

    def __init__(self, id: str, **kwargs: Any) -> None:
        """
        :id plot name, duplicate entry will update the plot

        :type the type of plot
        :caption caption of the plot
        :title title of the plot
        """

        # if we have an @ name we will override id and user
        if id[0] == "@":
            cand = id[1:].split("~")
            id = cand[1]
            uname = cand[0]
            kwargs["user"] = uname

        self.id = id

        self._vispath = "plots"
        self._type = "plot"

        self._freeze: bool = False

        # store pending updates when plot is frozen
        self._pending: Dict[str, str] = {}

        self.colors = NovemColors(self)
        self.cell = NovemCellConfig(self)
        self.config = NovemPlotConfig(self)

        super().__init__(**kwargs)

        self._parse_kwargs(**kwargs)

    def _set_data(self, data: Any, **kwargs: Any) -> Any:
        """
        Set's the data of the plot

        The parameter either needs to be a text string
        of CSV formatted text or an object with a to_csv
        function.
        """

        raw_str: str = ""
        to_csv = getattr(data, "to_csv", None)

        if callable(to_csv):
            # if data has a callable csv funciton we'll
            # assume it's a pandas dataframe
            raw_str = data.to_csv()
        else:
            raw_str = str(data)

        # invoke server write
        self._write("/data", raw_str)

        # also update our chart varibales
        self._parse_kwargs(**kwargs)

        # return a reference to the plot, from experience
        # after sending a dataframe to the plot the user
        # would rather operate on the plot object itself
        return self

    def __call__(self, data: Any, **kwargs: Any) -> Any:
        return self._set_data(data, **kwargs)

    def _parse_kwargs(self, **kwargs: Any) -> None:

        # first let our super do it's thing
        super()._parse_kwargs(**kwargs)

        # get a list of valid properties
        # exclude data as it needs to be run last
        props = [x for x in dir(self) if x[0] != "_" and x not in ["data", "read", "delete", "write"]]

        do_data = False
        for k, v in kwargs.items():
            if k == "data":
                do_data = True

            if k not in props:
                continue

            # print(f"{k} :: {v}")
            # set our value
            setattr(self, k, v)

        if do_data:
            data = kwargs["data"]
            setattr(self, "data", data)

    def _read(self, path: str) -> str:
        if self._freeze:
            if path in self._pending:
                return self._pending[path]

        return self.api_read(path)

    def _write(self, path: str, value: str) -> None:
        if self._freeze:
            self._pending[path] = value
        else:
            self.api_write(path, value)

    # we'll implement generic properties common across all plots here
    @property
    def type(self) -> str:
        return self._read("/config/type").strip()

    @type.setter
    def type(self, value: str) -> None:
        return self._write("/config/type", value)

    @property
    def name(self) -> str:
        return self._read("/name").strip()

    @name.setter
    def name(self, value: str) -> None:
        return self._write("/name", value)

    @property
    def description(self) -> str:
        return self._read("/description")

    @description.setter
    def description(self, value: str) -> None:
        return self._write("/description", value)

    @property
    def summary(self) -> str:
        return self._read("/summary")

    @summary.setter
    def summary(self, value: str) -> None:
        return self._write("/summary", value)

    @property
    def data(self) -> str:
        return self._read("/data")

    @data.setter
    def data(self, value: str) -> None:
        self._set_data(value)
        return None
        # return self._write("/data", value)

    @property
    def url(self) -> str:
        return self._read("/url").strip()

    @property
    def shortname(self) -> str:
        return self._read("/shortname").strip()

    ###
    # Interactive utility functions
    ###

    def df(self, data: Any, **kwargs: Any) -> Any:
        """
        Expects a dataframe as input and returns the
        same dataframe so it can be chained
        """
        self._set_data(data, **kwargs)
        return data

    # chainable utility function for setting values
    def w(self, key: str, value: str) -> Any:
        """
        Set a novem plot property, if key is a valid
        class porp then it will set that, else it will
        try to invoke an api call

        w('type','bar') -> invokes set attribute type
        w('/config/type','bar') -> invokes the api

        (both options results in the same effect)
        """
        props = [x for x in dir(self) if x[0] != "_" and x not in ["data", "read", "delete", "write"]]

        if key in props:
            self.__setattr__(key, value)
        else:
            self._write(key, value)

        return self

    # print our ansi version
    @property
    def x(self) -> None:
        print(self._read("/files/plot.ansi"))

        return None

    # print our img (png) special utility for qtconsole
    @property
    def i(self) -> Any:
        """
        Utility for getting a qtconsole image representation
        """
        from IPython.core.display import Image  # type: ignore

        return Image(
            self.api_read_bytes(f"/files/{self._type}.png"),
            retina=False,  # todo: lookup if computer supports retina and
            # request 2x resolution and set this variable
            width=900,  # todo: get img width from current session
        )

    ###
    # config variables
    ###

    # set the plot caption
    @property
    def caption(self) -> str:
        return self._read("/config/caption")

    @caption.setter
    def caption(self, value: str) -> None:
        return self._write("/config/caption", value)

    # set the plot title
    @property
    def title(self) -> str:
        return self._read("/config/title")

    @title.setter
    def title(self, value: str) -> None:
        return self._write("/config/title", value)

    ###
    # Deal with frozen plots
    ###

    def freeze(self) -> None:
        self._freeze = True

    def run(self) -> None:
        # push pending updates to server
        for path, value in self._pending.items():
            self.api_write(path, value)

        self._freeze = False

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "colors" and self.colors:
            self.colors.set(value)
        else:
            super().__setattr__(name, value)



import configparser
import io
import os
import platform
import select
import sys
import unicodedata
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from novem.types import Config

API_ROOT = "https://api.novem.no/v1/"
NOVEM_PATH = "novem"
NOVEM_NAME = "novem.conf"


class cl:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    FGGRAY = "\033[38;5;246m"
    BGGRAY = "\033[48;5;234m"


def disable_colors() -> None:
    c = cast(Any, cl)
    c.HEADER = ""
    c.OKBLUE = ""
    c.OKCYAN = ""
    c.OKGREEN = ""
    c.WARNING = ""
    c.FAIL = ""
    c.ENDC = ""
    c.BOLD = ""
    c.UNDERLINE = ""
    c.FGGRAY = ""
    c.BGGRAY = ""


def colors() -> None:
    # ignore color disable if --colors in argv
    for a in sys.argv:
        if os.name == "nt":
            from colorama import just_fix_windows_console  # type: ignore

            just_fix_windows_console()
        if a == "--color":
            return

    if os.name == "nt":
        # TODO: do some proper color detection on nt
        from colorama import just_fix_windows_console

        just_fix_windows_console()
        return

    # disable colors if not supported
    for handle in [sys.stdout, sys.stderr]:
        if (hasattr(handle, "isatty") and handle.isatty()) or ("TERM" in os.environ and os.environ["TERM"] == "ANSI"):
            if platform.system() == "Windows" and not ("TERM" in os.environ and os.environ["TERM"] == "ANSI"):
                disable_colors()
        else:
            disable_colors()


def get_user_config_directory() -> Union[str, None]:
    """Returns a platform-specific root directory for user config settings."""
    # On Windows, prefer %LOCALAPPDATA%, then %APPDATA%, since we can expect
    # the AppData directories to be ACLed to be visible only to the user and
    # admin users (https://stackoverflow.com/a/7617601/1179226). If neither is
    # set, return None instead of falling back to something that may be
    # world-readable.
    if os.name == "nt":
        appdata = os.getenv("LOCALAPPDATA")
        if appdata:
            return appdata
        appdata = os.getenv("APPDATA")
        if appdata:
            return appdata
        return None
    # On non-windows, use XDG_CONFIG_HOME if set, else default to ~/.config.
    xdg_config_home = os.getenv("XDG_CONFIG_HOME")
    if xdg_config_home:
        return xdg_config_home
    return os.path.join(os.path.expanduser("~"), ".config")


def get_config_path() -> Tuple[str, str]:
    """
    Get default configuration path
    """

    config_path: Union[str, None] = get_user_config_directory()
    novem_dir = f"{config_path}/{NOVEM_PATH}"
    novem_config = f"{config_path}/{NOVEM_PATH}/{NOVEM_NAME}"

    return (novem_dir, novem_config)


def get_current_config(
    **kwargs: Any,
) -> Tuple[bool, Config]:
    """
    Resolve and return the current config options

    Contains :
    current user
    current profile
    current token
    current api_root
    """

    co = Config(
        {
            "token": kwargs.get("token", None),
            "api_root": kwargs.get("api_root", API_ROOT),
            "ignore_ssl_warn": kwargs.get("ignore_ssl", False),
        }
    )

    if kwargs.get("token", False) or "ignore_config" in kwargs:
        return True, co

    # config path can be supplied as an option, if it is use that
    if "config_path" not in kwargs or not kwargs["config_path"]:
        (novem_dir, config_path) = get_config_path()
    else:
        config_path = kwargs["config_path"]

    config = configparser.ConfigParser()
    config.read(config_path)

    # the configuration file has an invalid format
    try:
        general = config["general"]
        profile = general["profile"]
        if "api_root" in general:
            co["api_root"] = general["api_root"]

    except KeyError:
        return (False, co)

    # override profile
    profile = kwargs.get("profile") or profile

    # get our config
    try:
        uc = config[f"profile:{profile}"]
        if "api_root" in uc:
            co["api_root"] = uc["api_root"]

        co["token"] = uc["token"]
        co["username"] = uc["username"]

        if "ignore_ssl_warn" in uc:
            co["ignore_ssl_warn"] = uc.getboolean("ignore_ssl_warn")

    except KeyError:
        return (True, co)

    # kwargs supercedes
    if kwargs.get("api_root", False):
        co["api_root"] = kwargs["api_root"]

    if kwargs.get("token", False):
        co["token"] = kwargs["token"]

    co["profile"] = profile
    return (True, co)


def pretty_format(values: List[Dict[str, str]], order: List[Dict[str, Any]]) -> str:
    """
    Constructs a pretty print table of the values in values
    in the order of List
    """

    colors()

    # lets' get total terminal width (we use 120 as default)
    try:
        (col, row) = os.get_terminal_size()
    except OSError:
        col = 120

    col = col - 2

    # padding width
    pw = 2

    # unicode aware string length https://stackoverflow.com/questions/33351599/
    def ucl(word: str) -> int:
        if not word:
            return 0
        return sum(1 for ch in word if unicodedata.combining(ch) == 0)

    # construct width map
    wm = {}
    for o in order:
        k = o["key"]
        try:
            cand = max([ucl(x[k]) for x in values])
        except ValueError:
            cand = 0

        wm[k] = max([cand, len(o["header"])])

    # let's calculate our actual widths
    if sum(wm.values()) + (len(order) - 1) * pw > col:
        # we need to adjust our sizing
        ainst = {}

        nts = 0
        for o in order:
            aos = wm[o["key"]]
            if o["overflow"] == "keep" and nts + aos < col:
                ainst[o["key"]] = "keep"
                nts += aos

        rem = col - nts - ((len(order) - 1) * pw)

        # allocate truncate width
        nks = [x for x in order if x["overflow"] != "keep"]
        for o in nks:
            wm[o["key"]] = int(rem / len(nks))

    # construct output string
    los = f"{cl.BOLD}"
    for o in order:
        w = f':<{wm[o["key"]]}'
        fmt = "{0" + w + "}"
        los += fmt.format(o["header"]) + " " * pw

    los += f"{cl.ENDC}\n"
    # sep
    for o in order:
        w = f':<{wm[o["key"]]}'
        fmt = "{0" + w + "}"
        # los += fmt.format("┄" * wm[o["key"]]) + " " * pw
        los += fmt.format("╌" * wm[o["key"]]) + " " * pw

    los += "\n"

    i = 0
    for p in values:
        for o in order:
            w = f':<{wm[o["key"]]}'
            fmt = "{0" + w + "}"
            vs = wm[o["key"]]
            ov = p[o["key"]]
            if ov is None:
                ov = ""

            if len(ov) > vs:
                val = ov[0 : vs - 3] + "..."
            else:
                val = ov[0:vs]

            if "fmt" in o:
                val = o["fmt"](val)

            val = fmt.format(val)

            if "clr" in o:
                if i % 2 == 0:
                    val = f'{o["clr"]}{val}{cl.ENDC}{cl.BGGRAY}'
                else:
                    val = f'{o["clr"]}{val}{cl.ENDC}'

            if o == order[-1]:
                pad = ""
            else:
                pad = " " * pw

            if i % 2 == 0:
                los += f"{cl.BGGRAY}" + val + pad + f"{cl.ENDC}"
            else:
                los += val + pad

        los += "\n"
        i += 1

    return los


def data_on_stdin() -> Optional[str]:
    """
    identify if there is data waiting on sys.stdin
    """

    @dataclass
    class Result:
        has_data: bool
        is_test: bool

    def _data_ready() -> Result:
        try:
            # use msvcrt on windows
            import msvcrt

            r = msvcrt.kbhit()  # type: ignore
            return Result(has_data=r, is_test=False)

        except ImportError:
            try:
                # use select on linux
                has_data = bool(
                    select.select(
                        [
                            sys.stdin,
                        ],
                        [],
                        [],
                        0.0,
                    )[0]
                )

                return Result(has_data=has_data, is_test=False)
            except io.UnsupportedOperation:
                # We're going to assume that this is the pytest wrapper
                # if sys.stdin is an instance of io.StringIO we are mocking data
                # on stdin, so it should be true. else ignore.
                has_data = isinstance(sys.stdin, io.StringIO)
                return Result(has_data=has_data, is_test=True)

    r = _data_ready()

    is_noninteractive = not sys.stdin.isatty()
    has_data = r.has_data or (is_noninteractive and not r.is_test)

    ctnt = "".join(sys.stdin.readlines()) if has_data else ""
    return ctnt if ctnt else None



"""
Config object for the standard plot structure
exposes the tree hierarchy with easy getters and setters
"""

from ..api_ref import NovemAPI


class NovemConfigBase(object):
    def __init__(self, api: NovemAPI) -> None:
        """ """
        self.api: NovemAPI = api

    def set(self, path: str, value: str) -> None:
        return self.api._write(path, value)

    def get(self, path: str) -> str:
        return self.api._read(path)


class NovemConfigLegend(NovemConfigBase):
    @property
    def format(self) -> str:
        return self.get("/config/legend/format")

    @format.setter
    def format(self, val: str) -> None:
        return self.set("/config/legend/format", val)

    @property
    def layout(self) -> str:
        return self.get("/config/legend/layout")

    @layout.setter
    def layout(self, val: str) -> None:
        return self.set("/config/legend/layout", val)

    @property
    def position(self) -> str:
        return self.get("/config/legend/position")

    @position.setter
    def position(self, val: str) -> None:
        return self.set("/config/legend/position", val)

    @property
    def type(self) -> str:
        return self.get("/config/legend/type")

    @type.setter
    def type(self, val: str) -> None:
        return self.set("/config/legend/type", val)


class NovemPlotConfig(object):
    def __init__(self, api: NovemAPI) -> None:
        """ """
        self.api: NovemAPI = api

        self.legend = NovemConfigLegend(api)



import datetime
import email.utils as eut
import json
import re
from typing import Any, Dict, List

from novem.exceptions import Novem404

from ..api_ref import NovemAPI
from ..utils import cl, colors, get_current_config, pretty_format


def list_vis(args: Dict[str, Any], type: str) -> None:
    colors()
    # get current plot list

    pfx = type[0].lower()

    novem = NovemAPI(**args)
    # see if list flag is set

    (config_status, config) = get_current_config(**args)

    plist = []

    usr = config["username"]
    if "for_user" in args and args["for_user"]:
        usr = args["for_user"]

    path = f"u/{usr}/{pfx}/"

    if "group" in args:
        # we're listing vis for a group
        query = ""
        group = args["group"]
        org = ""
        fu = ""

        if "for_user" in args and args["for_user"]:
            fu = args["for_user"]

        if "org" in args:
            org = args["org"]

        if group[0] in ["@", "+"]:
            query = group
        elif fu and group:
            query = f"@{usr}~{group}"
        elif org and group:
            query = f"+{org}~{group}"
        else:
            query = ""

        if query:
            path = f"o/{query}/{pfx}/"

    try:
        plist = json.loads(novem.read(path))
    except Novem404:
        plist = []

    if "filter" in args and args["filter"]:
        fv = args["filter"]
        if fv[0] != "^":
            fv = f".*{fv}"
        if fv[-1] != "$":
            fv = f"{fv}.*"

        flt = re.compile(fv, re.I)

        plist = [x for x in plist if (flt.match(x["id"]) or flt.match(x["name"]) or flt.match(x["type"]))]

    plist = sorted(plist, key=lambda x: x["id"])

    if args["list"]:

        # print to terminal
        for p in plist:
            print(p["id"])

        return

    ppo: List[Dict[str, Any]] = [
        {
            "key": "id",
            "header": f"{type} ID",
            "type": "text",
            "overflow": "keep",
        },
        # {
        #    "key": "id",
        #    "header": "ID",
        #    "type": "text",
        #    "overflow": "keep",
        # },
        {
            "key": "type",
            "header": "Type",
            "type": "text",
            "clr": cl.OKCYAN,
            "overflow": "keep",
        },
        {
            "key": "name",
            "header": "Name",
            "type": "text",
            "overflow": "truncate",
        },
        {
            "key": "uri",
            "header": "Url",
            "type": "url",
            "overflow": "keep",
        },
        {
            "key": "created",
            "header": "Created",
            "type": "date",
            "overflow": "keep",
        },
        {
            "key": "summary",
            "header": "Summary",
            "fmt": lambda x: x.replace("\n", ""),
            "type": "text",
            "overflow": "truncate",
        },
    ]

    for p in plist:
        nd = datetime.datetime(*eut.parsedate(p["created"])[:6])
        p["created"] = nd.strftime("%Y-%m-%d %H:%M")

    ppl = pretty_format(plist, ppo)

    print(ppl)

    return


def share_pretty_print(iplist: List[Dict[str, str]]) -> None:

    # modify our plist
    plist = []
    for p in iplist:
        if p["name"] == "public":
            p["summary"] = "Shared with the entire world"
            p["type"] = "special"
        if re.match("^@.+~.+$", p["name"]):
            p["summary"] = "Shared with all members of the given user group"
            p["type"] = "user group"
        if re.match("^\\+.+~.+$", p["name"]):
            p["summary"] = "Shared with all members of the given organisation group"
            p["type"] = "org group"

        plist.append(p)

    ppo: List[Dict[str, Any]] = [
        {
            "key": "name",
            "header": "Share Name",
            "type": "text",
            "overflow": "keep",
        },
        {
            "key": "type",
            "header": "Type",
            "type": "text",
            "clr": cl.OKCYAN,
            "overflow": "keep",
        },
        {
            "key": "created_on",
            "header": "Shared on",
            "type": "date",
            "overflow": "keep",
        },
        {
            "key": "summary",
            "header": "Summary",
            "fmt": lambda x: x.replace("\n", ""),
            "type": "text",
            "overflow": "truncate",
        },
    ]

    for p in plist:
        pds = eut.parsedate(p["created_on"])
        if pds:
            nd = datetime.datetime(*pds[:6])
            p["created_on"] = nd.strftime("%Y-%m-%d %H:%M")

    ppl = pretty_format(plist, ppo)
    print(ppl)


def list_vis_shares(vis_name: str, args: Dict[str, str], type: str) -> None:

    novem = NovemAPI(**args)
    # see if list flag is set

    pth = type.lower()

    (config_status, config) = get_current_config(**args)

    plist = []

    try:
        plist = json.loads(novem.read(f"vis/{pth}s/{vis_name}/shared"))
    except Novem404:
        plist = []

    if args["list"]:
        # print to terminal
        for p in plist:
            print(p["name"])
    else:
        share_pretty_print(plist)

    return
